"""
JSON based serialization
"""
#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2021- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

import jsonpickle
from ibmfl.message.message import Message
from ibmfl.message.serializer import Serializer


class JSONSerializer(Serializer):
    """
    Class for JSON Serializer
    """

    def serialize(self, msg):
        """
        Serialize a message using JSON

        :param msg: message to serialize
        :type msg: `Message`
        :return: serialize byte stream
        :rtype: `b[]`
        """
        msg_header = msg.get_header()
        data = msg.get_data()
        json_str_obj = jsonpickle.encode(
            {
                'header': msg_header,
                'data': data,
            })
        return json_str_obj.encode()  # return a byte stream

    def deserialize(self, serialized_byte_stream):
        """
        Deserialize a byte stream to a message

        :param serialized_byte_stream: byte stream
        :type serialized_byte_stream: `b[]`
        :return: deserialized message
        :rtype: `Message`
        """
        json_str_obj = serialized_byte_stream.decode()
        data_dict = jsonpickle.decode(json_str_obj)

        msg = Message(data=data_dict['data'])
        msg.set_header(data_dict['header'])

        return msg



#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2021- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

import abc
from enum import Enum


class FLConnection(abc.ABC):

    SENDER_STATUS = None
    RECEIVER_STATUS = None

    """
    Initialze sender and receiver

    :param kwargs: Dictionary of arguments required by each type of
        sender implementation
    """
    @abc.abstractmethod
    def initialize(self, **kwargs):
        pass

    """
    Validate config dictionary and initialize receiver

    :param kwargs: Dictionary of arguments required by each type of
        receiver implementation
    """
    @abc.abstractmethod
    def initialize_receiver(self, **kwargs):
        pass

    """
    Validate config dictionary and initialize sender

    :param kwargs: Dictionary of arguments required by each type of
        sender implementation
    """
    @abc.abstractmethod
    def initialize_sender(self, **kwargs):
        pass

    """
    Start connection
    """
    @abc.abstractmethod
    def start(self, **kwargs):
        pass

    """Stop and cleanup connection object
    """
    @abc.abstractmethod
    def stop(self, **kwargs):
        pass

    """Provide a connection information such that a node can communicate 
        to other nodes on how to communicate with it.
    """
    @abc.abstractmethod
    def get_connection_config(self):
        pass


class FLReceiver(abc.ABC):

    @abc.abstractmethod
    def initialize(self, **kwargs):
        pass

    """
    Start the server as per configuration after a clean intialize_server
    call
    :param kwargs: Dictionary of arguments required by each type of
        server implementation
    """
    @abc.abstractmethod
    def start(self, **kwargs):
        pass

    """
    Cleanup all the resources and stop serving
    :param kwargs: Dictionary of arguments required by each type of
        server implementation
    """
    @abc.abstractmethod
    def stop(self, **kwargs):
        pass


class FLSender(abc.ABC):

    @abc.abstractmethod
    def initialize(self, **kwargs):
        pass

    """
    Requests the end point as per the message forwarded to this module.
    Should have a logic to incorporate additional authentication details as
    per the configuration.

    :param kwargs: Dictionary of arguments required by each type of
        connection implementation
    """
    @abc.abstractmethod
    def send_message(self, **kwargs):
        pass

    """
    Cleanup the Sender object and close the hanging connections if any
    """
    @abc.abstractmethod
    def cleanup(self):
        pass


class ConnectionStatus(Enum):
    """Enum Class to describe the current status of the connection
    """
    INITIALIZED = 1
    STARTED = 2
    STOPPED = 3



#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2020- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

__all__ = [
    "WrongDeploymnetType",
    "ModelTypeNotSupported",
    "NotAutoAIExperiment",
    "EnvironmentNotSupported",
    'BatchJobFailed',
    'MissingScoringResults',
    'ModelStoringFailed',
    'DeploymentNotSupported',
    'MissingSpace',
    'PromotionFailed',
    'ModelPromotionFailed',
    'ServingNameNotAvailable'
]


from ibm_watson_machine_learning.utils import WMLClientError


class WrongDeploymnetType(WMLClientError, ValueError):
    def __init__(self, value_name, reason=None):
        WMLClientError.__init__(self, f"This deployment is not of type: {value_name} ", reason)


class ModelTypeNotSupported(WMLClientError, ValueError):
    def __init__(self, value_name, reason=None):
        WMLClientError.__init__(self, f"This model type is not supported yet: {value_name} ", reason)


class NotAutoAIExperiment(WMLClientError, ValueError):
    def __init__(self, value_name, reason=None):
        WMLClientError.__init__(self, f"This experiment_run_id is not from an AutoAI experiment: {value_name} ", reason)


class EnvironmentNotSupported(WMLClientError, ValueError):
    def __init__(self, value_name, reason=None):
        WMLClientError.__init__(self, f"This environment is not supported: {value_name}", reason)


class BatchJobFailed(WMLClientError, ValueError):
    def __init__(self, value_name=None, reason=None):
        WMLClientError.__init__(self, f"Batch job failed for job: {value_name}", reason)


class MissingScoringResults(WMLClientError, ValueError):
    def __init__(self, value_name=None, reason=None):
        WMLClientError.__init__(self, f"Scoring of deployment job: {value_name} not completed.", reason)


class ModelStoringFailed(WMLClientError, ValueError):
    def __init__(self, value_name=None, reason=None):
        WMLClientError.__init__(self, f"Model: {value_name} store failed.", reason)


class DeploymentNotSupported(WMLClientError, ValueError):
    def __init__(self, value_name=None, reason=None):
        WMLClientError.__init__(self, f"Deployment of type: {value_name} is not supported.", reason)


class MissingSpace(WMLClientError, ValueError):
    def __init__(self, value_name=None, reason=None):
        WMLClientError.__init__(self, f"Deployment needs to have space specified", reason)


class PromotionFailed(WMLClientError, ValueError):
    def __init__(self, project_id, space_id, promotion_response=None, reason=None):
        WMLClientError.__init__(self, f"Asset promotion from project {project_id} to space {space_id} failed."
                                      f" Full response {promotion_response}", reason)
        self.project_id = project_id
        self.space_id = space_id
        self.promotion_response = promotion_response
        self.reason = reason


class ModelPromotionFailed(PromotionFailed):
    def __init__(self, project_id, space_id, promotion_response=None, reason=None):
        PromotionFailed.__init__(self, project_id, space_id, promotion_response, reason)


class ServingNameNotAvailable(WMLClientError, ValueError):
    def __init__(self, message):
        WMLClientError.__init__(self, message)



#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2021- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

from __future__ import print_function
import logging
from ibmfl.exceptions import HyperparamsException

VALID_LOSSES = ['binary_crossentropy', 'categorical_crossentropy',
                'squared_error', 'least_absolute_deviation', 'auto']
VALID_SCORES = []

logger = logging.getLogger(__name__)


def validate_parameters(hyperparams):
    """
    Valdiates specific hyperparameters parameters that are defined in
    `hyperparams` variables. Raises an exception for certain parameters that
    are either required or beyond the defined range if they are defined.

    :param hyperparams: Dictionary containing values for the hyperparameters.
    :type  hyperparams: `dict`
    :rtype: None
    """
    try:
        # Check Hyperparameter Type
        if not isinstance(hyperparams, dict):
            raise ValueError('Provided hyperparameter is not valid.')

        # Global Hyperparameter Check
        if 'global' in hyperparams:
            params = hyperparams['global']
        else:
            raise ValueError('Global parameters have not been defined.')

        if 'learning_rate' in params:
            if params['learning_rate'] <= 0:
                raise ValueError('learning_rate={} must be strictly '
                                 'positive'.format(params['learning_rate']))
        else:
            raise ValueError('learning_rate has not been defined.')

        if 'loss' in params:
            if params['loss'] not in VALID_LOSSES:
                raise ValueError('Loss {} is currently not supported.'
                                 'Accpted losses: {}'.format(params['loss'],
                                                             ', '.join(VALID_LOSSES)))
        else:
            raise ValueError('loss has not been defined.')

        if 'num_classes' in params:
            # Validate Class Parameter Types
            if type(params['num_classes']) is not int:
                raise ValueError('Provided classes value\'s type is not valid, '
                                 'should be an int value >= 2 for classification.')

            # Validate Classification Case
            if params['loss'] != 'squared_error':
                if params['num_classes'] < 0:
                    raise ValueError('Provided class value must be >= 2 for '
                                     'classification.')

                if params['loss'] == 'binary_crossentropy' and \
                    params['num_classes'] != 2:
                    raise ValueError('Binary class models must have class of 2.')
                elif params['loss'] == 'categorical_crossentropy' and \
                    params['num_classes'] <= 2:
                    raise ValueError('Multiclass models must have class > 2.')
                elif params['loss'] == 'auto':
                    if params['num_classes'] < 2:
                        raise ValueError('Class value must be >= 2.')
                    else:
                        logging.warning(
                            'Obtaining class labels based on local dataset. '
                            'This may cause failures during aggregation '
                            'when parties have distinctive class labels.')
        else:
            # Handle Classes Not Defined Case
            if params['loss'] != 'squared_error':
                raise ValueError('Classes has not been defined. Should provide '
                                 'a value >= 2 for classification models.')

        if 'max_bins' in params:
            if not (2 <= params['max_bins'] and params['max_bins'] <= 255):
                raise ValueError('max_bins={} should be no smaller than 2 '
                                 'and no larger than 255.'.format(params['max_bins']))

        if 'max_iter' in params:
            if params['max_iter'] < 1:
                raise ValueError('max_iter={} must not be smaller '
                                 'than 1.'.format(params['max_iter']))
        else:
            raise ValueError('max_iter has not been defined.')

        if 'max_depth' in params:
            if params['max_depth'] is not None and params['max_depth'] < 0:
                raise ValueError('max_depth={} must be strictly greater'
                                 'than 0.'.format(params['max_depth']))

        if 'max_leaf_nodes' in params:
            if params['max_leaf_nodes'] is not None and params['max_leaf_nodes'] <= 1:
                raise ValueError('max_leaf_nodes={} must be strictly greater'
                                 'than 1.'.format(params['max_leaf_nodes']))

        if 'min_samples_leaf' in params:
            if params['min_samples_leaf'] is not None and params['min_samples_leaf'] < 0:
                raise ValueError('min_sample_leaf={} must not be smaller '
                                 'than 0'.format(params['min_samples_leaf']))

    except Exception as ex:
        logger.exception(str(ex))
        raise HyperparamsException('Defined global hyperparameters malformed. '+str(ex))


def init_parameters(obj, hyperparameters):
    """
    Initializes the hyperparameters to the corresponding object as an attribute.
    The parameters are propagated by reference, so changes should propagate to
    respective object passed into the function.

    :param obj: Any class object which to tie initialize the hyperparameters with.
    :type  obj: Object
    :param hyperparameters: Dictionary containing values for the hyperparameters.
    :type  hyperparameters: `dict`
    """
    # Initialize Global Configuration Parameter
    params = hyperparameters['global']
    setattr(obj, 'param', params)

    # Initialize Attributes (Pre-Checked Parameters)
    setattr(obj, 'learning_rate', params['learning_rate'])
    setattr(obj, 'loss', params['loss'])
    setattr(obj, 'max_iter', params['max_iter'])

    if params['loss'] == 'squared_error':
        setattr(obj, 'num_classes', 1)
    elif params['loss'] in ['binary_crossentropy', 'categorical_crossentropy', 'auto']:
        setattr(obj, 'num_classes', params['num_classes'])

    # Initialize Attributes (Optional Values - Based on Default Parameters)
    if 'l2_regularization' not in params or params['l2_regularization'] is None:
        setattr(obj, 'l2_regularization', 0)
    else:
        setattr(obj, 'l2_regularization', params['l2_regularization'])

    if 'max_bins' not in params:
        setattr(obj, 'max_bins', 255)
    else:
        setattr(obj, 'max_bins', params['max_bins'])

    if 'max_depth' not in params or params['max_depth'] is None:
        setattr(obj, 'max_depth', None)
    else:
        setattr(obj, 'max_depth', params['max_depth'])

    if 'max_leaf_nodes' not in params or params['max_leaf_nodes'] is None:
        setattr(obj, 'max_leaf_nodes', 31)
    else:
        setattr(obj, 'max_leaf_nodes', params['max_leaf_nodes'])

    if 'min_samples_leaf' not in params or params['min_samples_leaf'] is None:
        setattr(obj, 'min_samples_leaf', 20)
    else:
        setattr(obj, 'min_samples_leaf', params['min_samples_leaf'])

    if 'random_state' in params:
        setattr(obj, 'random_state', params['random_state'])
    else:
        setattr(obj, 'random_state', None)

    if 'scoring' in params:
        setattr(obj, 'scoring', params['scoring'])
    else:
        setattr(obj, 'scoring', None)

    if 'verbose' not in params or params['verbose'] is None:
        setattr(obj, 'verbose', False)
    else:
        setattr(obj, 'verbose', True)

    return obj



"""
This module will host all the exceptions which are raised by fl components
"""


#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2021- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

class FLException(Exception):
    pass


class DuplicateRouteException(FLException):
    pass


class InvalidConfigurationException(FLException):
    pass


class InvalidServerConfigurationException(InvalidConfigurationException):
    pass


class NotFoundException(FLException):
    pass


class LocalTrainingException(FLException):
    pass


class TCPMessageOutOfOrder(FLException):
    pass


class GlobalTrainingException(FLException):
    pass


class ModelException(FLException):
    pass


class ModelInitializationException(FLException):
    pass


class ModelUpdateException(FLException):
    pass


class HyperparamsException(FLException):
    pass


class CryptoKeyGenerationException(FLException):
    pass


class CryptoException(FLException):
    pass


class AuthorityException(FLException):
    pass


class KeyManagerException(FLException):
    pass


class WarmStartException(FLException):
    pass


class FusionException(FLException):
    pass


class QuorumException(FLException):
    pass



# coding: utf-8

#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2017- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

from pprint import pformat
from six import iteritems


class MlAssetsCreateModelInput(object):
    """
    NOTE: This class is auto generated by the swagger code generator program.
    Do not edit the class manually.
    """
    def __init__(self,
                 tags=None,
                 space=None,
                 project=None,
                 custom=None,
                 pipeline=None,
                 training_lib=None,
                 type=None,
                 domain=None,
                 name=None,
                 description=None,
                 transformed_label_column=None,
                 label_column=None,
                 training_data_references=None,
                 import_location=None,
                 size=None,
                 hyper_parameters=None,
                 runtime=None,
                 metrics=None,
                 schemas=None,
                 software_spec=None,
                 model_definition=None,
                 project_id=None,
                 space_id=None,
                 ):
        """
        MlAssetsCreateModelInput - a model defined in Swagger

        :param dict swaggerTypes: The key is attribute name
                                  and the value is attribute type.
        :param dict attributeMap: The key is attribute name
                                  and the value is json key in definition.
        """
        self.swagger_types = {
            'tags': 'list[TagRepository]',
            'space': 'ModelSpaces',
            'project': 'SpaceModels',
            'custom': 'ModelsCustom',
            'pipeline': 'PipelinesModels',
            'training_lib': 'TrainingModels',
            'type': 'str',
            'domain': 'str',
            'name': 'str',
            'description': 'str',
            'transformed_label_column': 'str',
            'label_column': 'str',
            'training_data_references': 'list[ModelContentLocation]',
            'import_location': 'ModelContentLocation',
            'size': 'ModelsSize',
            'hyper_parameters':'list[HyperParameters]',
            'runtime':'RuntimeModels',
            'metrics': 'ModelsMetrics',
            'schemas': 'ModelSchemas',
            'software_spec': 'SoftwareSpecModels',
            'model_definition': 'ModelDefinitionModels',
            'project_id': 'ProjectId',
            'space_id': 'SpaceId'

        }

        self.attribute_map = {
            'tags': 'tags',
            'space': 'space',
            'project': 'project',
            'custom': 'custom',
            'pipeline': 'pipeline',
            'training_lib': 'training_lib',
            'type': 'type',
            'domain': 'domain',
            'name': 'name',
            'description': 'description',
            'transformed_label_column': 'transformed_label_column',
            'label_column': 'label_column',
            'training_data_references': 'training_data_references',
            'import_location': 'import',
            'size': 'size',
            'hyper_parameters':'hyper_parameters',
            'runtime':'runtime',
            'metrics': 'metrics',
            'schemas': 'schemas',
            'software_spec': 'software_spec',
            'model_definition': 'model_definition',
            'project_id':'project_id',
            'space_id':'space_id'
        }

        self._tags = tags
        self._name = name
        self._description = description
        self._transformed_label_column = transformed_label_column
        self._label_column = label_column
        self._training_data_references = training_data_references
        self._schemas = schemas
        self._hyper_parameters = hyper_parameters
        self._runtime = runtime
        self._space = space
        self._project = project
        self._custom = custom
        self._pipeline =  pipeline
        self._training_lib =  training_lib
        self._type = type
        self._domain = domain
        self._import_location = import_location
        self._metrics = metrics
        self._size = size
        self._software_spec = software_spec
        self._model_definition = model_definition
        self._project_id = project_id
        self._space_id = space_id

    @property
    def tags(self):
        """
        Gets the tags of this ExperimentInputExperiments.


        :return: The tags of this ExperimentInputExperiments.
        :rtype: list[TagRepository]
        """
        return self._tags

    @tags.setter
    def tags(self, tags):
        """
        Sets the tags of this ExperimentInputExperiments.


        :param tags: The tags of this ExperimentInputExperiments.
        :type: list[TagRepository]
        """

        self._tags = tags

    @property
    def sofrware_spec(self):
        """
        Gets the software_spec of this Model


        :return: The software_spec of this models.
        :rtype: SoftwareSpecModels
        """
        return self._software_spec

    @sofrware_spec.setter
    def software_spec(self, software_spec):
        """
        Sets the software_spec of this Models.


        :param software_spec: The software_spec of this Models.
        :type: SoftwareSpecModels
        """

        self._software_spec = software_spec

    @property
    def model_definition(self):
        """
        Gets the model_definition of this Model


        :return: The model_definition of this models.
        :rtype: ModelDefinitionModels
        """
        return self._model_definition

    @model_definition.setter
    def model_definition(self, model_definition):
        """
        Sets the model_definition of this Model.


        :param model_definition: The model_definition of this Model.
        :type: ModelDefinitionModels
        """

        self._model_definition = model_definition

    @property
    def project_id(self):
        """
        Gets the project_id of this Model


        :return: The project id of this models.
        :rtype: str
        """
        return self._project_id

    @project_id.setter
    def project_id(self, project_id):
        """
        Sets the project_id of this Model input.


        :param project_id: The project_id of this Modelinput
        :type: str
        """

        self._project_id = project_id

    @property
    def space_id(self):
        """
        Gets the space_id of this Model


        :return: The space id of this models.
        :rtype: str
        """
        return self._space_id

    @space_id.setter
    def space_id(self, space_id):
        """
        Sets the space_id of this Model input.


        :param space_id: The space_id of this Modelinput
        :type: str
        """

        self._space_id = space_id

    @property
    def hyper_parameters(self):
        """
        Gets the hyper_parameters of this MlAssetsCreateModelInput.


        :return: The hyper_parameters of this MlAssetsCreateModelInput.
        :rtype: HyperParameters
        """
        return self._hyper_parameters

    @hyper_parameters.setter
    def hyper_parameters(self, hyper_parameters):
        """
        Sets the framework of this MlAssetsCreateModelInput.


        :param framework: The framework of this MlAssetsCreateModelInput.
        :type: HyperParameters
        """
        self._hyper_parameters = hyper_parameters

    @property
    def custom(self):
        """
        Gets the custom of this MlAssetsCreateModelInput.


        :return: The custom of this MlAssetsCreateModelInput.
        :rtype: custom
        """
        return self._custom

    @custom.setter
    def custom(self, custom):
        """
        Sets the framework of this MlAssetsCreateModelInput.


        :param framework: The framework of this MlAssetsCreateModelInput.
        :type: custom
        """
        self._custom = custom

    @property
    def schemas(self):
        """
        Gets the hyper_parameters of this MlAssetsCreateModelInput.


        :return: The hyper_parameters of this MlAssetsCreateModelInput.
        :rtype: HyperParameters
        """
        return self._schemas

    @schemas.setter
    def schemas(self, schemas):
        """
        Sets the framework of this MlAssetsCreateModelInput.


        :param framework: The framework of this MlAssetsCreateModelInput.
        :type: HyperParameters
        """
        self._schemas = schemas

    @property
    def name(self):
        """
        Gets the name of this MlAssetsCreateModelInput.


        :return: The name of this MlAssetsCreateModelInput.
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        """
        Sets the name of this MlAssetsCreateModelInput.


        :param name: The name of this MlAssetsCreateModelInput.
        :type: str
        """
        self._name = name

    @property
    def description(self):
        """
        Gets the description of this MlAssetsCreateModelInput.


        :return: The description of this MlAssetsCreateModelInput.
        :rtype: str
        """
        return self._description

    @description.setter
    def description(self, description):
        """
        Sets the description of this MlAssetsCreateModelInput.


        :param description: The description of this MlAssetsCreateModelInput.
        :type: str
        """
        self._description = description



    @property
    def transformed_label_column(self):
        """
        Gets the transformed_label_column of this MlAssetsCreateModelInput.


        :return: The transformed_label_column of this MlAssetsCreateModelInput.
        :rtype: str
        """
        return self._transformed_label_column

    @transformed_label_column.setter
    def transformed_label_column(self, transformed_label_column):
        """
        Sets the transformed_label_column of this MlAssetsCreateModelInput.


        :param transformed_label_column: The transformed_label_column of this MlAssetsCreateModelInput.
        :type: str
        """
        self._transformed_label_column = transformed_label_column



    @property
    def label_column(self):
        """
        Gets the label_column of this MlAssetsCreateModelInput.
        Name of the label column

        :return: The label_column of this MlAssetsCreateModelInput.
        :rtype: str
        """
        return self._label_column

    @label_column.setter
    def label_column(self, label_column):
        """
        Sets the label_column of this MlAssetsCreateModelInput.
        Name of the label column

        :param label_column: The label_column of this MlAssetsCreateModelInput.
        :type: str
        """
        self._label_column = label_column


    @property
    def training_data_references(self):
        """
        Gets the training_data of this MlAssetsCreateModelInput.
        :return: The training_data of this MlAssetsCreateModelInput.
        :rtype: ArrayDataInputRepository
        """
        return self._training_data_references


    @training_data_references.setter
    def training_data_reference(self, training_data_references):
        """
        Sets the training_data of this MlAssetsCreateModelInput.
        :param training_data: The training_data of this MlAssetsCreateModelInput.
        :type: ArrayDataInputRepository
        """
        self._training_data_references = training_data_references


    @property
    def evaluation(self):
        """
        Gets the evaluation of this MlAssetsCreateModelInput.
        Definition of the evaluation method used for pipeline models

        :return: The evaluation of this MlAssetsCreateModelInput.
        :rtype: EvaluationDefinitionRepository
        """
        return self._evaluation

    @evaluation.setter
    def evaluation(self, evaluation):
        """
        Sets the evaluation of this MlAssetsCreateModelInput.
        Definition of the evaluation method used for pipeline models

        :param evaluation: The evaluation of this MlAssetsCreateModelInput.
        :type: EvaluationDefinitionRepository
        """
        self._evaluation = evaluation


    @property
    def runtime(self):
        """
        Gets the runtime of this MlAssetsCreateModelInput.


        :return: The runtime of this MlAssetsCreateModelInput.
        :rtype: runtime
        """
        return self._runtime

    @runtime.setter
    def runtime(self, runtime):
        """
        Sets the runtime of this MlAssetsCreateModelInput.

        :param runtime: The runtime of this MlAssetsCreateModelInput.
        :type: runtime
        """
        self._runtime = runtime



    @property
    def space(self):
        """
        Gets the training_data of this MlAssetsCreateModelInput.
        :return: The training_data of this MlAssetsCreateModelInput.
        :rtype: ArrayDataInputRepository
        """
        return self._space

    @space.setter
    def space(self, space):
        """
        Sets the training_data of this MlAssetsCreateModelInput.
        :param training_data: The training_data of this MlAssetsCreateModelInput.
        :type: ArrayDataInputRepository
        """
        self._space = space

    @property
    def project(self):
        """
        Gets the project of this MlAssetsCreateModelInput.
        :return: The project of this MlAssetsCreateModelInput.
        :rtype: ArrayDataInputRepository
        """
        return self._project

    @project.setter
    def project(self, project):
        """
        Sets the project of this MlAssetsCreateModelInput.
        :param project: The training_data of this MlAssetsCreateModelInput.
        :type: ArrayDataInputRepository
        """
        self._project = project

    @property
    def pipeline(self):
        """
        Gets the training_data of this MlAssetsCreateModelInput.
        :return: The training_data of this MlAssetsCreateModelInput.
        :rtype: ArrayDataInputRepository
        """
        return self._pipeline

    @pipeline.setter
    def pipeline(self, pipeline):
        """
        Sets the training_data of this MlAssetsCreateModelInput.
        :param training_data: The training_data of this MlAssetsCreateModelInput.
        :type: ArrayDataInputRepository
        """
        self._pipeline = pipeline



    @property
    def training_lib(self):
        """
        Gets the training_data of this MlAssetsCreateModelInput.
        :return: The training_data of this MlAssetsCreateModelInput.
        :rtype: ArrayDataInputRepository
        """
        return self._training_lib

    @training_lib.setter
    def training_lib(self, training_lib):
        """
        Sets the training_data of this MlAssetsCreateModelInput.
        :param training_data: The training_data of this MlAssetsCreateModelInput.
        :type: ArrayDataInputRepository
        """
        self._training_lib = training_lib

    @property
    def type(self):
        """
        Gets the training_data of this MlAssetsCreateModelInput.
        :return: The training_data of this MlAssetsCreateModelInput.
        :rtype: ArrayDataInputRepository
        """
        return self._type

    @type.setter
    def type(self, type):
        """
        Sets the training_data of this MlAssetsCreateModelInput.
        :param training_data: The training_data of this MlAssetsCreateModelInput.
        :type: ArrayDataInputRepository
        """
        self._type = type

    @property
    def domain(self):
        """
        Gets the training_data of this MlAssetsCreateModelInput.
        :return: The training_data of this MlAssetsCreateModelInput.
        :rtype: ArrayDataInputRepository
        """
        return self._domain

    @domain.setter
    def domain(self, domain):
        """
        Sets the training_data of this MlAssetsCreateModelInput.
        :param training_data: The training_data of this MlAssetsCreateModelInput.
        :type: ArrayDataInputRepository
        """
        self._domain = domain

    @property
    def metrics(self):
        """
        Gets the training_data of this MlAssetsCreateModelInput.
        :return: The training_data of this MlAssetsCreateModelInput.
        :rtype: ArrayDataInputRepository
        """
        return self._metrics

    @metrics.setter
    def metrics(self, metrics):
        """
        Sets the training_data of this MlAssetsCreateModelInput.
        :param training_data: The training_data of this MlAssetsCreateModelInput.
        :type: ArrayDataInputRepository
        """
        self._metrics = metrics

    @property
    def import_location(self):
        """
        Gets the training_data of this MlAssetsCreateModelInput.
        :return: The training_data of this MlAssetsCreateModelInput.
        :rtype: ArrayDataInputRepository
        """
        return self._import_location

    @import_location.setter
    def import_location(self, import_location):
        """
        Sets the training_data of this MlAssetsCreateModelInput.
        :param training_data: The training_data of this MlAssetsCreateModelInput.
        :type: ArrayDataInputRepository
        """
        self._import_location = import_location

    @property
    def size(self):
        """
        Gets the training_data of this MlAssetsCreateModelInput.
        :return: The training_data of this MlAssetsCreateModelInput.
        :rtype: ArrayDataInputRepository
        """
        return self._size

    @size.setter
    def size(self, size):
        """
        Sets the training_data of this MlAssetsCreateModelInput.
        :param training_data: The training_data of this MlAssetsCreateModelInput.
        :type: ArrayDataInputRepository
        """
        self._size = size



    def to_dict(self):
        """
        Returns the model properties as a dict
        """
        result = {}

        for attr, _ in iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                if (attr == "import_location"):
                    result["import"] =  dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
                else:
                    result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                if (attr == "import_location"):
                    result["import"] = value
                else:
                    result[attr] = value

        return result

    def to_str(self):
        """
        Returns the string representation of the model
        """
        return pformat(self.to_dict())

    def __repr__(self):
        """
        For `print` and `pprint`
        """
        return self.to_str()

    def __eq__(self, other):
        """
        Returns true if both objects are equal
        """
        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """
        Returns true if both objects are not equal
        """
        return not self == other




#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2022- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------
from cryptography.exceptions import InvalidSignature
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import padding, rsa, utils
from cryptography.hazmat.primitives.serialization import load_pem_public_key

from ibmfl.crypto.crypto_exceptions import *
from ibmfl.crypto.infra.crypto_asym_int import CryptoAsym


class CryptoAsymRsa(CryptoAsym):
    """
    This class implements the interface for asymmetric encryption functions using RSA.
    """

    KEY_SIZE = 4096
    PUBLIC_EXPONENT = 65537
    CRYPTO_HASH = hashes.SHA256()
    SIGNATURE_BLOCK_SIZE = 1024 * 1024

    def __init__(self, key_file: str = None, password: bytes = None, **kwargs):
        super(CryptoAsymRsa, self).__init__(key_file, password)
        if key_file is None:
            self.private_key = CryptoAsymRsa.generate_key()
        else:
            self.private_key = self._read_key_file(key_file, password)

    def generate_key():
        private_key = rsa.generate_private_key(
            public_exponent=CryptoAsymRsa.PUBLIC_EXPONENT,
            key_size=CryptoAsymRsa.KEY_SIZE,
        )
        return private_key

    def get_public_key(self, type: str = "obj"):
        if self.private_key is None:
            raise KeyDistributionInputException("self.private_key is None")
        if type == "obj":
            ret = self.private_key.public_key()
        elif type == "pem":
            ret = self.private_key.public_key().public_bytes(
                encoding=serialization.Encoding.PEM, format=serialization.PublicFormat.SubjectPublicKeyInfo
            )
        else:
            raise KeyDistributionInputException("Invalid type=" + repr(type))
        return ret

    def _read_key_file(self, file_path: str, password):
        with open(file_path, "rb") as key_file:
            private_key = serialization.load_pem_private_key(
                key_file.read(),
                password=password,
            )
        return private_key

    def write_key_file(self, file_path: str, password: bytes):
        if self.private_key is None:
            raise KeyDistributionInputException("self.private_key is None")
        encryption_algorithm = serialization.NoEncryption()
        if password is not None:
            encryption_algorithm = serialization.BestAvailableEncryption(password)
        pem = self.private_key.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.PKCS8,
            encryption_algorithm=encryption_algorithm,
        )
        with open(file_path, "wb") as key_file:
            key_file.write(pem)
        return

    def encrypt(self, plain_data: bytes) -> bytes:
        if self.private_key is None:
            raise KeyDistributionInputException("self.private_key is None")
        cipher_data = self.private_key.public_key().encrypt(
            plain_data,
            padding.OAEP(
                mgf=padding.MGF1(algorithm=CryptoAsymRsa.CRYPTO_HASH), algorithm=CryptoAsymRsa.CRYPTO_HASH, label=None
            ),
        )
        return cipher_data

    def decrypt(self, cipher_data: bytes) -> bytes:
        if self.private_key is None:
            raise KeyDistributionInputException("self.private_key is None")
        plain_data = self.private_key.decrypt(
            cipher_data,
            padding.OAEP(
                mgf=padding.MGF1(algorithm=CryptoAsymRsa.CRYPTO_HASH), algorithm=CryptoAsymRsa.CRYPTO_HASH, label=None
            ),
        )
        return plain_data

    def encrypt_wkey(public_key, plain_data: bytes) -> bytes:
        if isinstance(public_key, rsa.RSAPublicKey):
            pb_key = public_key
        elif isinstance(public_key, bytes):
            pb_key = load_pem_public_key(public_key)
        else:
            raise KeyDistributionInputException("Invalid type of public_key=" + repr(type(public_key)))
        cipher_data = pb_key.encrypt(
            plain_data,
            padding.OAEP(
                mgf=padding.MGF1(algorithm=CryptoAsymRsa.CRYPTO_HASH), algorithm=CryptoAsymRsa.CRYPTO_HASH, label=None
            ),
        )
        return cipher_data

    def get_signature(self, data: bytes) -> bytes:
        if self.private_key is None:
            raise KeyDistributionInputException("self.private_key is None")
        hasher = hashes.Hash(CryptoAsymRsa.CRYPTO_HASH)
        n_blocks = int(len(data) / CryptoAsymRsa.SIGNATURE_BLOCK_SIZE)
        for i in range(n_blocks):
            hasher.update(data[i * CryptoAsymRsa.SIGNATURE_BLOCK_SIZE : (i + 1) * CryptoAsymRsa.SIGNATURE_BLOCK_SIZE])
        if len(data) % CryptoAsymRsa.SIGNATURE_BLOCK_SIZE > 0:
            hasher.update(data[n_blocks * CryptoAsymRsa.SIGNATURE_BLOCK_SIZE : len(data)])
        digest = hasher.finalize()
        signature = self.private_key.sign(
            digest,
            padding.PSS(mgf=padding.MGF1(CryptoAsymRsa.CRYPTO_HASH), salt_length=padding.PSS.MAX_LENGTH),
            utils.Prehashed(CryptoAsymRsa.CRYPTO_HASH),
        )
        return signature

    def verify_signature(public_key, signature: bytes, data: bytes) -> bool:
        if isinstance(public_key, rsa.RSAPublicKey):
            pb_key = public_key
        elif isinstance(public_key, bytes):
            pb_key = load_pem_public_key(public_key)
        else:
            raise KeyDistributionInputException("Invalid type of public_key=" + repr(type(public_key)))
        hasher = hashes.Hash(CryptoAsymRsa.CRYPTO_HASH)
        n_blocks = int(len(data) / CryptoAsymRsa.SIGNATURE_BLOCK_SIZE)
        for i in range(n_blocks):
            hasher.update(data[i * CryptoAsymRsa.SIGNATURE_BLOCK_SIZE : (i + 1) * CryptoAsymRsa.SIGNATURE_BLOCK_SIZE])
        if len(data) % CryptoAsymRsa.SIGNATURE_BLOCK_SIZE > 0:
            hasher.update(data[n_blocks * CryptoAsymRsa.SIGNATURE_BLOCK_SIZE : len(data)])
        digest = hasher.finalize()
        try:
            pb_key.verify(
                signature,
                digest,
                padding.PSS(mgf=padding.MGF1(CryptoAsymRsa.CRYPTO_HASH), salt_length=padding.PSS.MAX_LENGTH),
                utils.Prehashed(CryptoAsymRsa.CRYPTO_HASH),
            )
        except InvalidSignature:
            return False
        return True



#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2020- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

from ibm_watson_machine_learning.helpers.connections.connections import (
    DataConnection, S3Connection, S3Location, FSLocation, AssetLocation, CP4DAssetLocation,
    WMLSAssetLocation, CloudAssetLocation, WSDAssetLocation, DeploymentOutputAssetLocation,
    NFSConnection, NFSLocation, ConnectionAsset, ConnectionAssetLocation, DatabaseLocation, ContainerLocation)



#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2017- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

from .spark_pipeline_reader import SparkPipelineReader
from ibm_watson_machine_learning.libs.repo.mlrepository import MetaNames, MetaProps
from ibm_watson_machine_learning.libs.repo.mlrepository import PipelineArtifact
from .version_helper import VersionHelper
from ibm_watson_machine_learning.libs.repo.util.library_imports import LibraryChecker
from ibm_watson_machine_learning.libs.repo.base_constants import *

lib_checker = LibraryChecker()

if lib_checker.installed_libs[PYSPARK]:
    from pyspark.ml import Pipeline


class SparkPipelineArtifact(PipelineArtifact):
    """
    Class of pipeline artifacts created with MLRepositoryCLient.

    :param pyspark.ml.Pipeline ml_pipeline: Pipeline which will be wrapped

    :ivar pyspark.ml.Pipeline ml_pipeline: Pipeline associated with this artifact
    """
    def __init__(self, ml_pipeline, uid=None, name=None, meta_props=MetaProps({})):
        super(SparkPipelineArtifact, self).__init__(uid, name, meta_props)

        type_identified = False
        if lib_checker.installed_libs[PYSPARK]:
            if issubclass(type(ml_pipeline), Pipeline):
                type_identified = True

        if not type_identified and lib_checker.installed_libs[MLPIPELINE]:
            from mlpipelinepy.mlpipeline import MLPipeline
            if issubclass(type(ml_pipeline), MLPipeline):
                type_identified = True
        if not type_identified:
            raise ValueError('Invalid type for ml_pipeline: {}'.format(ml_pipeline.__class__.__name__))

        self.ml_pipeline = ml_pipeline
        self.meta.merge(
            MetaProps({
                MetaNames.FRAMEWORK_NAME: VersionHelper.pipeline_type(ml_pipeline),
                MetaNames.FRAMEWORK_VERSION: VersionHelper.getFrameworkVersion(ml_pipeline)
            })
        )

    def pipeline_instance(self):
        return self.ml_pipeline

    def reader(self):
        """
        Returns reader used for getting pipeline content.

        :return: reader for pyspark.ml.Pipeline
        :rtype: SparkPipelineReader
        """
        try:
            return self._reader
        except:
            self._reader = SparkPipelineReader(self.ml_pipeline, 'pipeline')
            return self._reader

    def _copy(self, uid):
        return SparkPipelineArtifact(self.ml_pipeline, uid, self.name, self.meta)



#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2022- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------
"""
Fully homomorphic encryption cryptosystem, built on helayer library.

* type:     public-key encryption
* setting:  Integer based

:Date: 2021
"""
import importlib
import logging

if importlib.util.find_spec("pyhelayers") is not None:
    import pyhelayers as pyhe

import numpy as np

from ibmfl.crypto.crypto_enumeration import CryptoEnum
from ibmfl.crypto.crypto_exceptions import CryptoException, KeyManagerException
from ibmfl.crypto.crypto_library import Cipher, Crypto
from ibmfl.model.model_update import ModelUpdate

logger = logging.getLogger(__name__)


class CryptoFHE(Crypto):
    """
    The crypto class for FHE crypto system, where its KeyGenerator and
    Cipher can be initialized.
    """

    def initialize(self, config):
        """
        Initializes cryptographic keys to generate the cipher object

        :param config: A dict that contains required configuration
        :type config: `dict`
        :return: None
        """
        logger.info("Initializing keys")
        try:
            key_config = {
                "type": CryptoEnum.CRYPTO_FHE.value,
            }
            keys = self.key_manager.initialize_keys(config=key_config)
        except KeyManagerException as ex:
            logger.exception(ex)
            raise KeyManagerException("Initialization of FHE context failed.")

        logger.info("Initializing a FHE Cipher")
        try:
            self.cipher = CipherFHE(keys, 0)
        except CryptoException as ex:
            logger.error(str(ex))
            raise CryptoException("Initialization of FHE Cipher failed.")

        self.private_fusion_weight = config.get("private_fusion_weights")
        if self.private_fusion_weight is None:
            logger.info("No flag for privacy of fusion weights in config. " "Setting to default value of False.")
            self.private_fusion_weight = False
        if not isinstance(self.private_fusion_weight, bool):
            raise CryptoException("Flag private_fusion_weights must be boolean.")

    def encrypt(self, model_update, key="weights", **kwargs):
        """
        Method for the encryption operations FL party will perform
        during training. It currently calls the `encrypt_weights` method
        defined inside the Cipher class.

        :param model_update: ModelUpdate object to be encrypted.
        :type model_update: `ModelUpdate`
        :param key: model
        :type key: `str`
        :return: The resulting ciphertext of model update.
        :type: `ModelUpdate`
        """
        if not isinstance(model_update, ModelUpdate):
            raise CryptoException("not a ModelUpdate instance to be encrypted")

        ct_model_update = ModelUpdate()
        weights = model_update.get(key)
        counts = kwargs.get("counts")
        if self.private_fusion_weight:
            ct_weights, ct_counts = self.cipher.encrypt_weights(weights, counts=counts)
            ct_model_update.add("ct_weights", ct_weights)
            ct_model_update.add("ct_counts", ct_counts)
        else:
            ct_weights = self.cipher.encrypt_weights(weights)
            ct_model_update.add("ct_weights", ct_weights)
            ct_model_update.add("counts", counts)
        ct_model_update.add("weights_dtype", self.cipher.weights_dtype)

        return ct_model_update

    def decrypt(self, ct_model_update, **kwargs):
        """
        Method for processing fused model update in ciphertext that
        FL party will perform during training.

        :param ct_model_update: A `ModelUpdate` object containing the provided
        ciphertext to be decrypted. The ciphertext should be of
        type `list` of `np.ndarray`.
        :type ct_model_update: `ModelUpdate`
        :return: A `ModelUpdate` object containing the resulting decrypted \
        plain text, where the plain text is of type `list` of `np.ndarray`.
        :rtype: `ModelUpdate`
        """
        ct_weights = ct_model_update.get("ct_weights")
        weights = self.cipher.decrypt_weights(ct_weights, weights_dtype=ct_model_update.get("weights_dtype"))
        res_model_update = ModelUpdate()
        res_model_update.add("weights", weights)

        return res_model_update

    def aggregate_collected_ciphertext_response(self, lst_model_updates, key="ct_weights", **kwargs):
        """
        Receives a list of `model update`, where a `model_update` is of type
        `ModelUpdate`, using the encrypted values (indicating by the `key`)
        included in each `model_update`, it returns the fused model udpate.
        NOTE: if `private_fusion_weight` is set to `False`, the fusion weight
        of aggregation (`lst_fusion_weights`) should provided in plaintext;
        otherwise, the fusion weight is protected by FHE and it is required to
        provide `min_counts` and `max_counts` for counted tranining samples.

        :param: lst_model_updates. List of `model_update` of type
            `ModelUpdate` including the encrypted values (indicating by
            the `key`)
            to be averaged.
        :type lst_model_updates:  `list`

        :return: results of fused encrypted model update
        :rtype: `ModelUpdate`
        """
        logger.info("Start ciphertext fusing")

        # private_fusion_weight defined as the trigger to protect counts (fusion weight).
        weights_dtype = None
        if not self.private_fusion_weight:
            lst_ct_weights = []
            lst_fusion_weights = []
            for model_update in lst_model_updates:
                lst_ct_weights.append(model_update.get("ct_weights"))
                if model_update.get("counts") is not None:
                    lst_fusion_weights.append(model_update.get("counts"))
                cur_weights_dtype = model_update.get("weights_dtype")
                if cur_weights_dtype is not None:
                    if weights_dtype is None:
                        weights_dtype = cur_weights_dtype
                    elif weights_dtype != cur_weights_dtype:
                        logger.warning(
                            "aggregate_collected_ciphertext_response: weights dtype mismatch: "
                            + str(weights_dtype)
                            + " != "
                            + str(cur_weights_dtype)
                        )
            if len(lst_fusion_weights) > 0:
                lst_fusion_weights /= np.sum(lst_fusion_weights)
            else:
                lst_fusion_weights = [1.0 / len(lst_model_updates) for _ in lst_model_updates]
            ct_fused_weights = self.cipher.fuse_weights_scalar(lst_ct_weights, lst_fusion_weights)
        else:
            lst_ct_weights = []
            lst_ct_counts = []
            for model_update in lst_model_updates:
                lst_ct_weights.append(model_update.get("ct_weights"))
                lst_ct_counts.append(model_update.get("ct_counts"))
                cur_weights_dtype = model_update.get("weights_dtype")
                if cur_weights_dtype is not None:
                    if weights_dtype is None:
                        weights_dtype = cur_weights_dtype
                    elif weights_dtype != cur_weights_dtype:
                        logger.warning(
                            "aggregate_collected_ciphertext_response: weights dtype mismatch: "
                            + str(weights_dtype)
                            + " != "
                            + str(cur_weights_dtype)
                        )
            low_bound = kwargs.get("min_counts", 0)
            up_bound = kwargs.get("max_counts", 1000)
            eps = kwargs.get("eps", 1e-5)
            ct_fused_weights = self.cipher.fuse_weights(lst_ct_weights, lst_ct_counts, low_bound, up_bound, eps)

        model_update = ModelUpdate(ct_weights=ct_fused_weights, weights_dtype=weights_dtype)
        logger.info("Ciphertext fusion done")
        return model_update


class CipherFHE(Cipher):
    """
    This class module implements the encryption and decryption functionality
    for general FHE crypto system assuming key(s) will be provided
    during initialization.
    """

    def __init__(self, keys, precision, **kwargs):
        """
        Initializes a crypto cipher for FHE crypto system with
        provided crypto keys.

        :param keys: Provided crypto keys
        :type keys: `dict`
        :return: None
        """
        if "files" in keys:
            self.context = pyhe.DefaultContext()
            self.context.load_from_file(keys["files"]["context"])
            if "key" in keys["files"]:
                self.context.load_secret_key_from_file(keys["files"]["key"])
                self.crypto = pyhe.TTEncoder(self.context)
                self.slots = self.context.slot_count()
        elif "distribution" in keys:
            self.context = None
            self.crypto = None
        else:
            self.context = pyhe.DefaultContext()
            self.context.load_from_buffer(keys["pp"])
            if "sk" in keys:
                self.context.load_secret_key(keys["sk"])
                self.crypto = pyhe.TTEncoder(self.context)
                self.slots = self.context.slot_count()
        self.weights_dtype = None

    def set_keys(self, keys):
        logger.debug("CipherFHE set_keys start")
        self.context = pyhe.DefaultContext()
        self.context.load_from_buffer(keys["pp"])
        if "sk" in keys:
            self.context.load_secret_key(keys["sk"])
            self.crypto = pyhe.TTEncoder(self.context)
            self.slots = self.context.slot_count()
        logger.debug("CipherFHE set_keys end")

    def encrypt_weights(self, model_weights, **kwargs):
        """
        Encrypts the provided plain text of type list of `np.ndarray`.

        :param model_weights: The provided plain text to be encrypted.
        :type model_weights: `list` of `np.ndarray`
        :return: The resulting ciphertext of the provided plain text.
        :type: `list` of `np.ndarray`
        """
        logger.debug("CipherFHE encrypt_weights start")
        cipher_weights = []
        shape = pyhe.TTShape([1, self.slots])
        for i in range(len(model_weights)):
            weight = model_weights[i]
            if not isinstance(weight, np.ndarray):
                weight = np.asarray(weight)
            ct = self.crypto.encode_encrypt(shape, weight.reshape((1, weight.size)))
            cipher_weights.append([ct.save_to_buffer(), weight.shape])
            if self.weights_dtype is None:
                self.weights_dtype = weight.dtype
            elif self.weights_dtype != weight.dtype:
                logger.warning(
                    "encrypt_weights: weights dtype for multiple layers is different: "
                    "[self.weights_dtype=" + str(self.weights_dtype) + "] [weight.dtype=" + str(weight.dtype) + "]"
                )

        # protect counts for FedAvg fusion method
        counts = kwargs.get("counts", None)
        if counts is not None:
            shape_counts = shape.get_with_duplicated_dims([0, 1])
            ct_counts = self.crypto.encode_encrypt(shape_counts, np.array(counts).reshape(1, 1))
            return cipher_weights, ct_counts.save_to_buffer()
        logger.debug("CipherFHE encrypt_weights end")
        return cipher_weights

    def decrypt_weights(self, ct_model_weights, **kwargs):
        """
        Decrypts the provided ciphertext of type list of `np.ndarray`.

        :param ct_model_weights: The provided ciphertext to be decrypted.
        :type ct_model_weights: `list` of `np.ndarray`
        :return: The resulting decrypted plain text.
        :rtype: `list` of `np.ndarray`
        """
        logger.debug("CipherFHE decrypt_weights start")

        if self.weights_dtype is None:
            weights_dtype = kwargs.get("weights_dtype")
            if weights_dtype is not None:
                self.weights_dtype = weights_dtype
                logger.info("decrypt_weights: weights dtype obtained from incoming model: " + str(self.weights_dtype))
            else:
                self.weights_dtype = np.zeros(1, np.float32).dtype
                logger.warning(
                    "decrypt_weights: weights dtype is None. " "Using default dtype: " + str(self.weights_dtype)
                )

        dec_weights = list()
        for i in range(len(ct_model_weights)):
            target_shape = ct_model_weights[i][1]
            ct = pyhe.CTileTensor(self.context)
            ct.load_from_buffer(ct_model_weights[i][0])
            dec_weight = self.crypto.decrypt_decode_double(ct)
            if dec_weight.dtype != self.weights_dtype:
                dec_weight = dec_weight.astype(self.weights_dtype)
            dec_weights.append(dec_weight.reshape(target_shape))
        logger.debug("CipherFHE decrypt_weights end")
        return dec_weights

    def fuse_weights_scalar(self, ct_lst_weights, lst_fusion_weight):
        """
        Fuse a set of encrypted model weights with plaintext fusion
        weight for each model.

        :param ct_lst_weights: the list of ciphertext model weights
        :type ct_lst_weights: `list` of `np.ndarray`
        :param lst_fusion_weight: the list of fusion weight in plaintext
        :type lst_fusion_weight: `list`
        :return: fused ciphertext weights
        """
        logger.debug("CipherFHE fuse_weights_scalar start")
        if len(lst_fusion_weight) != len(ct_lst_weights):
            raise CryptoException("Size of fusion weight does not match" "the size of recevied ciphertext list")
        fused_ct_keras = ct_lst_weights[0]
        for layer_idx in range(len(fused_ct_keras)):
            fused_ct = pyhe.CTileTensor(self.context)
            fused_ct.load_from_buffer(fused_ct_keras[layer_idx][0])
            fused_ct.multiply_scalar(lst_fusion_weight[0])
            for party_idx in range(1, len(ct_lst_weights)):
                party_ct = pyhe.CTileTensor(self.context)
                party_ct.load_from_buffer(ct_lst_weights[party_idx][layer_idx][0])
                party_ct.multiply_scalar(lst_fusion_weight[party_idx])
                fused_ct.add(party_ct)
            fused_ct_keras[layer_idx][0] = fused_ct.save_to_buffer()
        logger.debug("CipherFHE fuse_weights_scalar end")
        return fused_ct_keras

    def fuse_weights(self, ct_lst_weights, ct_lst_fusions, low_bound, up_bound, eps):
        """
        Fuse a set of encrypted model weights with plaintext fusion
        weight for each model.

        :param ct_lst_weights: the list of ciphertext model weights
        :type ct_lst_weights: `list` of `np.ndarray`
        :param ct_lst_fusions: the list of encrypted fusion weight in plaintext
        :type ct_lst_fusions: `list` of `np.ndarray`
        :param low_bound: low bound for the input for the inverse function
        :type low_bound: `int`
        :param up_bound: up bound for the input for the inverse function
        :type up_bound: `int`
        :param eps: eps parameter
        :type eps: `double`
        :return: inversed encrypted fusion weights
        :rtype: `list` of `np.ndarray`
        """
        logger.debug("CipherFHE fuse_weights start")
        # address the encrypted fusion weight aggregation with eps
        logger.info("perform inverse of encrypted aggregated fusion weight with eps")
        ct_fusion_sum = pyhe.CTileTensor(self.context)
        ct_fusion_sum.load_from_buffer(ct_lst_fusions[0])
        for party_idx in range(1, len(ct_lst_fusions)):
            ct_fusion_party = pyhe.CTileTensor(self.context)
            ct_fusion_party.load_from_buffer(ct_lst_fusions[party_idx])
            ct_fusion_sum.add(ct_fusion_party)
        ct_fusion_sum.add_scalar(eps)
        func_eval = pyhe.TTFunctionEvaluator(self.context)
        ct_inverse = func_eval.inverse(ct_fusion_sum, low_bound, up_bound)

        # address weighted fusion over encryptd fusions and models
        logger.info("perform weighted sum over encrypted model weights")
        fused_ct_model = ct_lst_weights[0]
        ct_fusion_party0 = pyhe.CTileTensor(self.context)
        ct_fusion_party0.load_from_buffer(ct_lst_fusions[0])
        for layer_idx in range(len(fused_ct_model)):
            fused_ct = pyhe.CTileTensor(self.context)
            fused_ct.load_from_buffer(fused_ct_model[layer_idx][0])
            fused_ct.multiply(ct_fusion_party0)
            for party_idx in range(1, len(ct_lst_weights)):
                party_ct = pyhe.CTileTensor(self.context)
                party_ct.load_from_buffer(ct_lst_weights[party_idx][layer_idx][0])
                party_ct_fusion = pyhe.CTileTensor(self.context)
                party_ct_fusion.load_from_buffer(ct_lst_fusions[party_idx])
                party_ct.multiply(party_ct_fusion)
                fused_ct.add(party_ct)
            fused_ct.multiply(ct_inverse)
            fused_ct_model[layer_idx][0] = fused_ct.save_to_buffer()
        logger.debug("CipherFHE fuse_weights end")
        return fused_ct_model



#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2023- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------
import logging
from typing import Any, List, Mapping, Optional

try:
    from langchain.llms.base import LLM
    from langchain.llms.utils import enforce_stop_tokens
except ImportError:
    raise ImportError("Could not import langchain: Please install langchain extension.")
from ibm_watson_machine_learning.foundation_models import Model

logger = logging.getLogger(__name__)


class WatsonxLLM(LLM):
    """
    `LangChain CustomLLM <https://python.langchain.com/docs/modules/model_io/models/llms/custom_llm>`_ wrapper for watsonx foundation models.

    :param model: foundation model inference object instance
    :type model: Model

    **Supported chain types:**
        * `LLMChain`,
        * `TransformChain`,
        * `SequentialChain`,
        * `SimpleSequentialChain`
        * `ConversationChain` (including `ConversationBufferMemory`)
        * `LLMMathChain` (``bigscience/mt0-xxl``, ``eleutherai/gpt-neox-20b``, ``ibm/mpt-7b-instruct2``, ``bigcode/starcoder``, ``meta-llama/llama-2-70b-chat`` models only)
        * `LLMBashChain` (``eleutherai/gpt-neox-20b`` model only)

    **Instantiate the WatsonxLLM interface**

    .. code-block:: python

        from ibm_watson_machine_learning.foundation_models import Model
        from ibm_watson_machine_learning.metanames import GenTextParamsMetaNames as GenParams
        from ibm_watson_machine_learning.foundation_models.extensions.langchain import WatsonxLLM

        generate_params = {
            GenParams.MAX_NEW_TOKENS: 25
        }

        model = Model(
            model_id="google/flan-ul2",
            credentials={
                "apikey": "***",
                "url": "https://us-south.ml.cloud.ibm.com"
            },
            params=generate_params,
            project_id="*****"
        )

        custom_llm = WatsonxLLM(model=model)

    """
    model: Model = None
    llm_type: str = "IBM watsonx.ai"

    def __init__(self, model: Model) -> None:
        super(WatsonxLLM, self).__init__()
        self.model = model

    @property
    def _identifying_params(self) -> Mapping[str, Any]:
        """Get the identifying parameters."""
        return {
            **{"model_id": self.model.model_id},
            **{"params": self.model.params},
            **{"project_id": self.model.project_id},
            **{"space_id": self.model.space_id},
        }

    @property
    def _llm_type(self) -> str:
        """Return type of llm."""
        return self.llm_type

    def _call(self, prompt: str, stop: Optional[List[str]] = None) -> str:
        """Call the IBM watsonx.ai inference endpoint.
        Args:
            prompt: The prompt to pass into the model.
            stop: Optional list of stop words to use when generating.
        Returns:
            The string generated by the model.
        Example:
            .. code-block:: python
                model = Model(
                    model_id="google/flan-ul2",
                    credentials={
                        "apikey": "***",
                        "url": "https://us-south.ml.cloud.ibm.com"
                        },
                    project_id="*****"
                    )
                llm = WatsonxLLM(model=model)
                response = llm("What is a molecule")
        """
        text = self.model.generate_text(prompt=prompt)
        logger.info("Output of watsonx.ai call: {}".format(text))
        if stop is not None:
            text = enforce_stop_tokens(text, stop)
        return text



#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2017- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

import tarfile, gzip, shutil


class CompressionUtil(object):
    @staticmethod
    def create_tar(path, arcname, archive_path):
        tar = tarfile.open(archive_path, "w")
        tar.add(path, arcname)
        tar.close()

    @staticmethod
    def extract_tar(archive_path, path):
        tar = tarfile.open(archive_path)
        tar.extractall(path)
        tar.close()

    @staticmethod
    def compress_file_gzip(filepath, gzip_filepath):
        with open(filepath, 'rb') as f_in, gzip.open(gzip_filepath, 'wb+') as f_out:
            shutil.copyfileobj(f_in, f_out)

    @staticmethod
    def decompress_file_gzip(gzip_filepath, filepath):
        with gzip.open(gzip_filepath, 'rb') as f:
            content = f.read()
            output_f = open(filepath, 'wb+')
            output_f.write(content)
            output_f.close()



#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2017- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

import inspect
import sys

from py4j.java_collections import MapConverter, ListConverter
from pyspark import SparkContext

from Wrapper import PythonJavaConversions

class DAG:
    """
    The base class for creating a DAG object.

    A DAG is a graph data structure in which each stage is an ETL transformer which
    transforms one dataframe into another.
    The start node of a DAG is referred to as root and the end nodes are referred to as leaves.
    A DAG can have one or more stages in the root as well as in the leaves.
    """
    _sc = SparkContext._active_spark_context
    _logSrcLang = "Py:"

    def __init__(self):
        self._jLogger = self._sc._jvm.org.apache.log4j.Logger
        self.logger = self._jLogger.getLogger(self.__class__.__name__)
        methodname = str(inspect.stack()[0][3])
        logMsg = self._logSrcLang + self.__class__.__name__ + ":" + methodname + ": [Params: None]"
        self.logger.info(logMsg)

        self._jDAGObject = None
        self._jPipelineWrapper = self._sc._jvm.com.ibm.analytics.ngp.pipeline.pythonbinding.Pipelines
        self._to_list = self._jPipelineWrapper.toList
        self._to_tuple = self._jPipelineWrapper.toTuple2
        self._to_map = self._jPipelineWrapper.toScalaMap

    def _to_java_stage(self, stage):
        methodname = str(inspect.stack()[0][3])
        logMsg = self._logSrcLang + self.__class__.__name__ + ":" + methodname + ": [Params: " + " stage => " + str(
                stage) + " ]"
        self.logger.info(logMsg)

        if hasattr(stage, "_to_java"):
            jstage = stage._to_java()
        else:
            jstage = PythonJavaConversions._to_java_stage(stage)
        return jstage

    def __isvalidtype(self, stage, label=None):
        methodname = str(inspect.stack()[1][3])

        if not (hasattr(stage, 'transform') or hasattr(stage, 'fit') or
                hasattr(stage, 'run')):
            logMsg = self._logSrcLang + self.__class__.__name__ + ":" + methodname + ": Stage assigned to a DAG " \
                    "should be a Transformer, Estimator or a Source, got " + str(type(stage)) + \
                    ". Refer the API documentation for syntax."
            self.logger.error(logMsg)
            raise TypeError("Stage assigned to a DAG should be a Transformer, Estimator or a Source, "
                            "got %s" % type(stage))

        if label and not isinstance(label, basestring):
            logMsg = self._logSrcLang + self.__class__.__name__ + ":" + methodname + ": Label assigned to a DAG " \
                    "should be a basestring, got " + str(type(label)) + ". Refer the API documentation for syntax."
            self.logger.error(logMsg)
            raise TypeError("Label assigned to a DAG should be a basestring, got %s" % type(label))

        else:
            return True

    def start(self, stages):
        """
        Associates stage(s) to the start of the DAG.

        :param stages: stages can be

        - A single transformer

        - List of transformers

        - A single transformer and a label specified as a tuple. The label has to be specified
          only when the Pipeline needs to be saved and loaded later.

        - List of transformers with labels, where each element of the list is a tuple consisting of
          transformer and a label. The label has to be specified only when the Pipeline needs to be
          saved and loaded later.

        :return: A :class:`DAG` object.
        """
        methodname = str(inspect.stack()[0][3])
        logMsg = self._logSrcLang + self.__class__.__name__ + ":" + methodname + ": [Params: " + " stages => " + str(
                stages) + " ]"
        self.logger.info(logMsg)

        if type(stages) is list:
            pjstages = []
            isTuple = 0
            for element in stages:
                if type(element) is tuple :
                    isTuple=1
                    pstage, label = element
                    if self.__isvalidtype(pstage,label):
                        jstage = self._to_java_stage(pstage)
                        pjstages.append((jstage, label))
                else:
                    pstage = element
                    if self.__isvalidtype(pstage):
                        jstage = self._to_java_stage(pstage)
                        pjstages.append(jstage)
            if isTuple :
                jlist = ListConverter().convert(pjstages, self._sc._gateway._gateway_client)
                stuple = self._jPipelineWrapper.toScalaTupleFromJavaList(jlist)
                self._jDAGObject = self._jPipelineWrapper.getDAG().start(stuple,
                                                                         self._jPipelineWrapper.getDummyImplicit())
            else:
                jstages = self._to_list(ListConverter().convert(pjstages, self._sc._gateway._gateway_client))
                self._jDAGObject = self._jPipelineWrapper.getDAG().start(jstages)

        elif type(stages) is tuple:
            pstage, label = stages
            pjstages=[]
            if self.__isvalidtype(pstage,label):
                jstage = self._to_java_stage(pstage)
                pjstages.append(jstage)
                pjstages.append(label)
                jlist = ListConverter().convert(pjstages, self._sc._gateway._gateway_client)
                stuple = self._jPipelineWrapper.toTuple2(jlist)
                self._jDAGObject = self._jPipelineWrapper.getDAG().start(stuple)

        else:
            if self.__isvalidtype(stages):
                jstage = self._to_java_stage(stages)
                self._jDAGObject = self._jPipelineWrapper.getDAG().start(jstage)

        return self

    def append(self, stage, childstage):
        """
        Appends a child stage to an existing stage of a DAG node. The Transformer
        specified by `childstage` is applied on the DataFrame after `stage`.

        :param stage: An existing Transformer stage of a DAG
        :param childstage: A new Transformer stage to be appended to `stage`
        :return: A :class:`DAG` object with `childstage` appended to `stage`
        """
        methodname = str(inspect.stack()[0][3])
        logMsg = self._logSrcLang + self.__class__.__name__ + ":" + methodname + ": [Params: " + " stage => " + str(
                stage) + " childstage => " + str(childstage) + " ]"
        self.logger.info(logMsg)

        if self.__isvalidtype(stage) and self.__isvalidtype(childstage):
            self._jDAGObject = self._jDAGObject.append(self._to_java_stage(stage), self._to_java_stage(childstage))
            return self

    def merge(self, parentstagetuple, stage):
        """
        Merges two Transformer stage outputs of a DAG to form a single stage. The DataFrames
        from the Transformer stages are passed to `stage` Transformer, which produces a single
        DataFrame as output.

        :param parentstagetuple: A tuple specifying the Transformer stages of the DAG to be merged.
        :param stage: A Transformer stage, which takes the outputs from two DataFrames and merges.
        :return: A :class:`DAG` object
        """
        methodname = str(inspect.stack()[0][3])
        logMsg = self._logSrcLang + self.__class__.__name__ + ":" + methodname + ": [Params: " + " parentstagetuple => " \
                "" + str(parentstagetuple) +  " stage => " + str(stage) + " ]"
        self.logger.info(logMsg)

        if type(parentstagetuple) is not tuple:
            logMsg = self._logSrcLang + self.__class__.__name__ + ":" + methodname + ": Parent stages for merge " \
                    "should be a tuple, got " + str(type(parentstagetuple)) + ". Refer the API " \
                    "documentation for syntax."
            self.logger.error(logMsg)
            raise TypeError("Parent stages for merge should be a tuple, got %s" % type(parentstagetuple))

        parentstage1, parentstage2 = parentstagetuple
        if self.__isvalidtype(stage):
            jstage = self._to_java_stage(stage)

        if self.__isvalidtype(parentstage1) and self.__isvalidtype(parentstage2):
            jparentstage1 = self._to_java_stage(parentstage1)
            jparentstage2 = self._to_java_stage(parentstage2)
            # Convert the tuple parentstages2 to list ,since java does not have tuple
            pjparentStages = [jparentstage1, jparentstage2]

        # call _to_tuple to convert java list to scala tuple2
        jparentstages = self._to_tuple(
                ListConverter().convert(pjparentStages, self._sc._gateway._gateway_client))
        self._jDAGObject = self._jDAGObject.merge(jparentstages, jstage)
        return self

    def setBindings(self, binddict):
        """
        Sets the input DataFrame(s) for the root stage(s) of the DAG.
        
        :param binddict: A dictionary specifying the input DataFrame(s) to root stage(s) mapping.
        :return: A :class:`DAG` object
        """
        methodname = str(inspect.stack()[0][3])
        logMsg = self._logSrcLang + self.__class__.__name__ + ":" + methodname + ": [Params: " + " binddict => " + str(
                binddict) + " ]"
        self.logger.info(logMsg)

        if type(binddict) is not dict:
            logMsg = self._logSrcLang + self.__class__.__name__ + ":" + methodname + ": Stage and Input " \
                    "DataFrame for a DAG should be specified as a dictionary, got " \
                    + str(type(binddict))
            self.logger.error(logMsg)
            raise TypeError("Stage and Input DataFrame for a DAG should be specifed as a dictionary, got %s"
                            % type(binddict))
        
        partialJbinds = {}
        for k, v in binddict.iteritems():
            partialJbinds[self._to_java_stage(k)] = v._jdf

        jbinds = MapConverter().convert(partialJbinds, self._sc._gateway._gateway_client)
        sca_map = self._to_map(jbinds)
        self._jDAGObject = self._jDAGObject.setBindings(sca_map)
        return self

    def fork(self, stage, childstages):
        """
        Forks the DAG to form child stages after stage. The `childstages` here specifies a
        list of Transformers.  If the list has more than one Transformer, each Transformer
        specified in `childstages` is applied separately on the dataframe after `stage` resulting
        in more than one execution paths.

        :param stage: An existing Transformer stage of a DAG
        :param childstages: A list of Transformer stages to be forked after stage.
        :return: A :class:`DAG` object with `childstages` forked after `stage`
        """
        methodname = str(inspect.stack()[0][3])
        logMsg = self._logSrcLang + self.__class__.__name__ + ":" + methodname + ": [Params: " + " stage => " + str(
                stage) + " childstages => " + str(childstages) + " ]"
        self.logger.info(logMsg)

        if self.__isvalidtype(stage):
            pjChildStages = []
            for pstage in childstages:
                if self.__isvalidtype(pstage):
                    pjChildStages.append(self._to_java_stage(pstage))

        jChildStages = self._to_list(ListConverter().convert(pjChildStages, self._sc._gateway._gateway_client))
        self._jDAGObject = self._jDAGObject.fork(self._to_java_stage(stage), jChildStages)

        return self

    def _to_java(self):

        methodname = str(inspect.stack()[0][3])
        logMsg = self._logSrcLang + self.__class__.__name__ + ":" + methodname + ": [Params: None]"
        self.logger.info(logMsg)

        return self._jDAGObject



# coding: utf-8

#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2018- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

from pprint import pformat
from six import iteritems
import re


class TrainingReferenceExperiments(object):
    """
    NOTE: This class is auto generated by the swagger code generator program.
    Do not edit the class manually.
    """
    def __init__(self, name=None, training_definition_url=None, command=None, hyper_parameters_optimization=None, compute_configuration=None, pretrained_model_url=None):
        """
        TrainingReferenceExperiments - a model defined in Swagger

        :param dict swaggerTypes: The key is attribute name
                                  and the value is attribute type.
        :param dict attributeMap: The key is attribute name
                                  and the value is json key in definition.
        """
        self.swagger_types = {
            'name': 'str',
            'training_definition_url': 'str',
            'command': 'str',
            'hyper_parameters_optimization': 'HyperParametersOptimizationExperiments',
            'compute_configuration': 'ComputeConfigurationExperiments',
            'pretrained_model_url': 'str'
        }

        self.attribute_map = {
            'name': 'name',
            'training_definition_url': 'training_definition_url',
            'command': 'command',
            'hyper_parameters_optimization': 'hyper_parameters_optimization',
            'compute_configuration': 'compute_configuration',
            'pretrained_model_url': 'pretrained_model_url'
        }

        self._name = name
        self._training_definition_url = training_definition_url
        self._command = command
        self._hyper_parameters_optimization = hyper_parameters_optimization
        self._compute_configuration = compute_configuration
        self._pretrained_model_url = pretrained_model_url

    @property
    def name(self):
        """
        Gets the name of this TrainingReferenceExperiments.


        :return: The name of this TrainingReferenceExperiments.
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        """
        Sets the name of this TrainingReferenceExperiments.


        :param name: The name of this TrainingReferenceExperiments.
        :type: str
        """

        self._name = name

    @property
    def training_definition_url(self):
        """
        Gets the training_definition_url of this TrainingReferenceExperiments.


        :return: The training_definition_url of this TrainingReferenceExperiments.
        :rtype: str
        """
        return self._training_definition_url

    @training_definition_url.setter
    def training_definition_url(self, training_definition_url):
        """
        Sets the training_definition_url of this TrainingReferenceExperiments.


        :param training_definition_url: The training_definition_url of this TrainingReferenceExperiments.
        :type: str
        """

        self._training_definition_url = training_definition_url

    @property
    def command(self):
        """
        Gets the command of this TrainingReferenceExperiments.
        only required for DL - it is overwriting what we have in {baseUrl}/v3/ml_assets/training_definitions/{id}

        :return: The command of this TrainingReferenceExperiments.
        :rtype: str
        """
        return self._command

    @command.setter
    def command(self, command):
        """
        Sets the command of this TrainingReferenceExperiments.
        only required for DL - it is overwriting what we have in {baseUrl}/v3/ml_assets/training_definitions/{id}

        :param command: The command of this TrainingReferenceExperiments.
        :type: str
        """

        self._command = command

    @property
    def hyper_parameters_optimization(self):
        """
        Gets the hyper_parameters_optimization of this TrainingReferenceExperiments.
        It is not used for model_definition

        :return: The hyper_parameters_optimization of this TrainingReferenceExperiments.
        :rtype: HyperParametersOptimizationExperiments
        """
        return self._hyper_parameters_optimization

    @hyper_parameters_optimization.setter
    def hyper_parameters_optimization(self, hyper_parameters_optimization):
        """
        Sets the hyper_parameters_optimization of this TrainingReferenceExperiments.
        It is not used for model_definition

        :param hyper_parameters_optimization: The hyper_parameters_optimization of this TrainingReferenceExperiments.
        :type: HyperParametersOptimizationExperiments
        """

        self._hyper_parameters_optimization = hyper_parameters_optimization

    @property
    def compute_configuration(self):
        """
        Gets the compute_configuration of this TrainingReferenceExperiments.


        :return: The compute_configuration of this TrainingReferenceExperiments.
        :rtype: ComputeConfigurationExperiments
        """
        return self._compute_configuration

    @compute_configuration.setter
    def compute_configuration(self, compute_configuration):
        """
        Sets the compute_configuration of this TrainingReferenceExperiments.


        :param compute_configuration: The compute_configuration of this TrainingReferenceExperiments.
        :type: ComputeConfigurationExperiments
        """

        self._compute_configuration = compute_configuration

    @property
    def pretrained_model_url(self):
        """
        Gets the pretrained_model_url of this TrainingReferenceExperiments.
        This is URL to the model which is stored in the repository service

        :return: The pretrained_model_url of this TrainingReferenceExperiments.
        :rtype: str
        """
        return self._pretrained_model_url

    @pretrained_model_url.setter
    def pretrained_model_url(self, pretrained_model_url):
        """
        Sets the pretrained_model_url of this TrainingReferenceExperiments.
        This is URL to the model which is stored in the repository service

        :param pretrained_model_url: The pretrained_model_url of this TrainingReferenceExperiments.
        :type: str
        """

        self._pretrained_model_url = pretrained_model_url

    def to_dict(self):
        """
        Returns the model properties as a dict
        """
        result = {}

        for attr, _ in iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """
        Returns the string representation of the model
        """
        return pformat(self.to_dict())

    def __repr__(self):
        """
        For `print` and `pprint`
        """
        return self.to_str()

    def __eq__(self, other):
        """
        Returns true if both objects are equal
        """
        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """
        Returns true if both objects are not equal
        """
        return not self == other



#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2020- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

from ibm_watson_machine_learning.experiment.autoai.optimizers.local_auto_pipelines import LocalAutoPipelines
from ibm_watson_machine_learning.experiment.autoai.optimizers.remote_auto_pipelines import RemoteAutoPipelines



#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2017- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

from .spark_artifact_loader import SparkArtifactLoader


class SparkPipelineModelLoader(SparkArtifactLoader):
    """
        Returns pipeline model instance associated with this model artifact.

        :return: pipeline model
        :rtype: pyspark.ml.PipelineModel
        """
    def model_instance(self):
        return self.load()


#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2017- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

class ArtifactReader(object):
    """
    Template reader class used to read artifacts.
    """

    def read(self):
        """
        Returns stream object with content of pipeline/pipeline model.

        :return: binary stream
        """
        pass

    def close(self):
        """
        Closes stream to content.
        """
        pass



#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

from typing import Optional

from ibm_watson_machine_learning.wml_client_error import WMLClientError
from ibm_watson_machine_learning.messages.messages import Messages

class CPDVersion:
    """Storage for cpd version. Comparison operators are 
    overloaded to allow comparison with numeric values. 

    Class attribute:
    :param supported_version_list: List of supported CPD versions.
    :type supported_version_list: list

    Attribute:
    :param cpd_version: Store CPD version.
    :type cpd_version: Optional[str], optional

    .. code-block:: python

        from ibm_watson_machine_learning.utils import CPDVersion

        version = CPDVersion()

        if not version:
            print("CPD version is None")

        version.cpd_version = '3.5'

        if version > 3:
            print("version greater than 3.0")

    """    
    supported_version_list = ['3.5', '4.0', '4.5', '4.6', '4.7', '4.8']
    
    def __init__(self, version: Optional[str] = None):
        self.cpd_version = version

    def __str__(self):
        version = self.__cpd_version
        return f"CPD version {version}" if version is not None else ""

    @property
    def cpd_version(self):
        """Attribute that stores cpd version. Before the value is set,
          validation is performed against a supported versions.
        """        
        return self.__cpd_version
    
    @cpd_version.setter
    def cpd_version(self, value):
        if value is None:
           self.__cpd_version = value 
        elif str(value) in CPDVersion.supported_version_list:
            self.__cpd_version = str(value)
        else:
            raise WMLClientError(Messages.get_message(', '.join(CPDVersion.supported_version_list),
                                                       message_id="invalid_version"))
        
    @cpd_version.getter
    def cpd_version(self):
        value = self.__cpd_version
        if value is None:
            return None
        else:
            dot_index = value.find(".")
            value = value[:dot_index+1] + value[dot_index+1:].replace(".","")
            return float(value)

    def __bool__(self):
        return bool(self.cpd_version)

    def __eq__(self, value):
        return self.cpd_version.__eq__(value)
        
    def __ne__(self, value):
        return not self.__eq__(value)

    def __lt__(self, value):
        return bool(self) and self.cpd_version.__lt__(value)

    def __le__(self, value):
        return bool(self) and (self.__lt__(value) or self.__eq__(value))

    def __gt__(self, value):
        return bool(self) and not self.__le__(value)

    def __ge__(self, value):
        return bool(self) and not self.__lt__(value)


#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2017- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

import json
from collections import namedtuple


class Json2ObjectMapper(object):
    @staticmethod
    def read(json_holder):  # TODO map
        if json_holder is None:
            return {}
        elif isinstance(json_holder, str):
            return json.loads(json_holder)
        else:
            return json_holder  # TODO json.loads(json_str, object_hook=lambda d: namedtuple('X', d.keys())(*d.values()))

    @staticmethod
    def to_dict(json_str):
        if isinstance(json_str, str):
            return json.loads(json_str)
        else:
            raise ValueError('Incorrect type')

    @staticmethod
    def to_object(json_holder):
        if isinstance(json_holder, str):
            return json.loads(json_holder, object_hook=lambda d: namedtuple('X', d.keys())(*d.values()))
        elif isinstance(json_holder, object):
            return json_holder
        else:
            raise ValueError('Incorrect type')



#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2020- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

__all__ = [
    'get_project',
    'get_wmls_credentials_and_space_ids'
]


import base64
import json
from typing import Any, Dict, Tuple, List, TYPE_CHECKING

import ibm_watson_machine_learning._wrappers.requests as requests

from ibm_watson_machine_learning.href_definitions import IAM_TOKEN_API, HrefDefinitions
from ibm_watson_machine_learning.wml_client_error import ApiRequestFailure
from .errors import NotInWatsonStudio, CredentialsNotFound

if TYPE_CHECKING:
    from ibm_watson_machine_learning.workspace import WorkSpace


def get_project() -> Any:
    """Try to import project_lib and get user corresponding project."""
    try:
        from project_lib import Project

    except ModuleNotFoundError:
        raise NotInWatsonStudio(reason="You are not in Watson Studio or Watson Studio Desktop environment. "
                                       "Cannot access to project metadata.")

    try:
        access = Project.access()

    except RuntimeError:
        raise CredentialsNotFound(reason="Your WSD environment does not have correctly configured "
                                         "connection to WML Server or you are not in WSD environment. "
                                         "In that case, please provide WMLS credentials and space_id.")

    return access


def get_wmls_credentials_and_space_ids(local_path: str) -> Tuple[List[Dict], List[str]]:
    """Parse project.json file and get WMLS credentials associated with WSD project.

    :param local_path: path to project.json file with project configuration in WSD
    :type local_path: str

    :return: two lists with WMLS credentials and corresponding space_ids
    :rtype: tuple[list[dict], list[str]]
    """
    try:
        with open(local_path, 'r') as f:
            data = json.load(f)

        credentials = [instance['credentials'] for instance in data['compute']]
        for instance in credentials:
            instance['version'] = "2.0"
            instance['instance_id'] = "wml_local"
            instance['password'] = base64.decodebytes(bytes(instance['password'].encode())).decode()

        space_ids = [instance['properties']['space_guid'] for instance in data['compute']]

    except (FileNotFoundError, KeyError):
        raise CredentialsNotFound(reason="Your WSD environment does not have correctly configured "
                                         "connection to WML Server or you are not in WSD environment. "
                                         "In that case, please provide WMLS credentials and space_id.")

    return credentials, space_ids



#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2017- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

from pprint import pformat
from six import iteritems
import re

class MlAssetsGetPreSignedUrlOutput(object):
    def __init__(self, contentUrl=None):
        self.swagger_types = {
            'contentUrl': 'str'
        }

        self.attribute_map = {
            'contentUrl': 'contentUrl'
        }

        self._contentUrl = contentUrl
    @property
    def contentUrl(self):

        return self._contentUrl

    @contentUrl.setter
    def contentUrl(self, contentUrl):

        self._contentUrl = contentUrl

    def to_dict(self):
        """
        Returns the model properties as a dict
        """
        result = {}

        for attr, _ in iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """
        Returns the string representation of the model
        """
        return pformat(self.to_dict())

    def __repr__(self):
        """
        For `print` and `pprint`
        """
        return self.to_str()

    def __eq__(self, other):
        """
        Returns true if both objects are equal
        """
        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """
        Returns true if both objects are not equal
        """
        return not self == other


#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2021- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------
"""
 An enumeration class for the message type field which describe what
 kind of data is being sent inside the Message
"""
from enum import Enum

__author__ = "Supriyo Chakraborty, Shalisha Witherspoon, Dean Steuer"


class MessageType(Enum):
    """
    Message types for communication between party and aggregator
    """

    MODEL_UPDATE = 1
    MODEL_HYPERPARAMETERS = 2
    MODEL_PARAMETERS = 3
    REQUEST_MODEL_HYPERPARAMETERS = 4
    REQUEST_MODEL_UPDATE = 5
    REGISTER = 6
    TRAIN = 7
    SAVE_MODEL = 8
    PREDICT_MODEL = 9
    EVAL_MODEL = 10
    ACK = 11
    SYNC_MODEL = 12
    STOP = 14
    PREPROCESS = 15
    REQUEST_CERT = 16
    GENERATE_KEYS = 17
    DISTRIBUTE_KEYS = 18
    SET_KEYS = 19
    ERROR_AUTH = 400



#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2020- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

from abc import ABC, abstractmethod

__all__ = [
    "BaseExperiment"
]


class BaseExperiment(ABC):
    """Base abstract class for Experiment."""

    @abstractmethod
    def runs(self, *, filter: str):
        """Get the historical runs but with WML Pipeline name filter.

        :param filter: WML Pipeline name to filter the historical runs
        :type filter: str
        """
        pass



# coding: utf-8

#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2017- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

from pprint import pformat
from six import iteritems
import re


class ModelVersionOutputEntity(object):
    """
    NOTE: This class is auto generated by the swagger code generator program.
    Do not edit the class manually.
    """
    def __init__(self, model=None, content_href=None, training_data_ref=None, evaluation=None, current_metrics=None,
                 content_status=None, content_location=None,hyper_parameters=None):
        """
        ModelVersionOutputEntity - a model defined in Swagger

        :param dict swaggerTypes: The key is attribute name
                                  and the value is attribute type.
        :param dict attributeMap: The key is attribute name
                                  and the value is json key in definition.
        """
        self.swagger_types = {
            'model': 'ModelVersionOutputEntityModel',
            'content_href': 'str',
            'training_data_ref': 'ModelTrainingDataRef',
            'evaluation': 'EvaluationDefinition',
            'current_metrics': 'ModelMetrics',
            'content_status': 'ContentStatus',
            'content_location': 'ContentLocation',
            'hyper_parameters' : 'list[HyperParameters]'
        }

        self.attribute_map = {
            'model': 'model',
            'content_href': 'contentHref',
            'training_data_ref': 'trainingDataRef',
            'evaluation': 'evaluation',
            'current_metrics': 'currentMetrics',
            'content_status': 'contentStatus',
            'content_location': 'contentLocation',
            'hyper_parameters': 'hyper_parameters'
        }

        self._model = model
        self._content_href = content_href
        self._training_data_ref = training_data_ref
        self._evaluation = evaluation
        self._current_metrics = current_metrics
        self._content_status = content_status
        self._content_location = content_location
        self._hyper_parameters = hyper_parameters

    @property
    def model(self):
        """
        Gets the model of this ModelVersionOutputEntity.


        :return: The model of this ModelVersionOutputEntity.
        :rtype: ModelVersionOutputEntityModel
        """
        return self._model

    @model.setter
    def model(self, model):
        """
        Sets the model of this ModelVersionOutputEntity.


        :param model: The model of this ModelVersionOutputEntity.
        :type: ModelVersionOutputEntityModel
        """

        self._model = model

    @property
    def hyper_parameters(self):
        """
        Gets the hyper_parameters of this MlAssetsCreateModelInput.


        :return: The hyper_parameters of this MlAssetsCreateModelInput.
        :rtype: HyperParameters
        """
        return self._hyper_parameters

    @hyper_parameters.setter
    def hyper_parameters(self, hyper_parameters):
        """
        Sets the framework of this MlAssetsCreateModelInput.


        :param framework: The framework of this MlAssetsCreateModelInput.
        :type: HyperParameters
        """
        self._hyper_parameters = hyper_parameters

    @property
    def content_href(self):
        """
        Gets the content_href of this ModelVersionOutputEntity.
        Href to the binary content of the model version. Make GET on it to download this file

        :return: The content_href of this ModelVersionOutputEntity.
        :rtype: str
        """
        return self._content_href

    @content_href.setter
    def content_href(self, content_href):
        """
        Sets the content_href of this ModelVersionOutputEntity.
        Href to the binary content of the model version. Make GET on it to download this file

        :param content_href: The content_href of this ModelVersionOutputEntity.
        :type: str
        """

        self._content_href = content_href

    @property
    def training_data_ref(self):
        """
        Gets the training_data_ref of this ModelVersionOutputEntity.


        :return: The training_data_ref of this ModelVersionOutputEntity.
        :rtype: ModelTrainingDataRef
        """
        return self._training_data_ref

    @training_data_ref.setter
    def training_data_ref(self, training_data_ref):
        """
        Sets the training_data_ref of this ModelVersionOutputEntity.


        :param training_data_ref: The training_data_ref of this ModelVersionOutputEntity.
        :type: ModelTrainingDataRef
        """

        self._training_data_ref = training_data_ref

    @property
    def evaluation(self):
        """
        Gets the evaluation of this ModelVersionOutputEntity.
        Definition of the evaluation method used for pipeline models

        :return: The evaluation of this ModelVersionOutputEntity.
        :rtype: EvaluationDefinition
        """
        return self._evaluation

    @evaluation.setter
    def evaluation(self, evaluation):
        """
        Sets the evaluation of this ModelVersionOutputEntity.
        Definition of the evaluation method used for pipeline models

        :param evaluation: The evaluation of this ModelVersionOutputEntity.
        :type: EvaluationDefinition
        """

        self._evaluation = evaluation

    @property
    def current_metrics(self):
        """
        Gets the current_metrics of this ModelVersionOutputEntity.


        :return: The current_metrics of this ModelVersionOutputEntity.
        :rtype: ModelMetrics
        """
        return self._current_metrics

    @current_metrics.setter
    def current_metrics(self, current_metrics):
        """
        Sets the current_metrics of this ModelVersionOutputEntity.


        :param current_metrics: The current_metrics of this ModelVersionOutputEntity.
        :type: ModelMetrics
        """

        self._current_metrics = current_metrics

    @property
    def content_status(self):
        """
        Gets the content_status of this ModelVersionOutputEntity.


        :return: The content_status of this ModelVersionOutputEntity.
        :rtype: ContentStatus
        """
        return self._content_status

    @content_status.setter
    def content_status(self, content_status):
        """
        Sets the content_status of this ModelVersionOutputEntity.


        :param content_status: The content_status of this ModelVersionOutputEntity.
        :type: ContentStatus
        """
        self._content_status = content_status

    @property
    def content_location(self):
        """
        Gets the content_status of this ModelVersionOutputEntity.


        :return: The content_status of this ModelVersionOutputEntity.
        :rtype: ContentLocation
        """
        return self._content_location

    @content_location.setter
    def content_location(self, content_location):
        """
        Sets the content_location of this ModelVersionOutputEntity.


        :param content_location: The content_status of this ModelVersionOutputEntity.
        :type: ContentLocation
        """
        self._content_status = content_location

    def to_dict(self):
        """
        Returns the model properties as a dict
        """
        result = {}

        for attr, _ in iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """
        Returns the string representation of the model
        """
        return pformat(self.to_dict())

    def __repr__(self):
        """
        For `print` and `pprint`
        """
        return self.to_str()

    def __eq__(self, other):
        """
        Returns true if both objects are equal
        """
        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """
        Returns true if both objects are not equal
        """
        return not self == other



#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2018- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

import logging, re, os

from ibm_watson_machine_learning.libs.repo.mlrepository import MetaNames
from ibm_watson_machine_learning.libs.repo.mlrepositoryartifact.libraries_artifact import LibrariesArtifact
from ibm_watson_machine_learning.libs.repo.swagger_client.api_client import ApiException
import json
from .libraries_adapter import WmlLibrariesAdapter
from ibm_watson_machine_learning.libs.repo.swagger_client.models import LibrariesDefinitionInput, LibrariesDefinitionInputPlatform, PatchOperationLibraries


logger = logging.getLogger('LibrariesCollection')


class LibrariesCollection:
    """
    Client operating on libraries in repository service.

    :param str base_path: base url to Watson Machine Learning instance
    :param MLRepositoryApi repository_api: client connecting to repository rest api
    :param MLRepositoryClient client: high level client used for simplification and argument for constructors
    """
    def __init__(self, base_path, repository_api, client):

        from ibm_watson_machine_learning.libs.repo.mlrepositoryclient import MLRepositoryClient
        from ibm_watson_machine_learning.libs.repo.mlrepositoryclient import MLRepositoryApi

        if not isinstance(base_path, str) and not isinstance(base_path, unicode):
            raise ValueError('Invalid type for base_path: {}'.format(base_path.__class__.__name__))

        if not isinstance(repository_api, MLRepositoryApi):
            raise ValueError('Invalid type for repository_api: {}'.format(repository_api.__class__.__name__))

        if not isinstance(client, MLRepositoryClient):
            raise ValueError('Invalid type for client: {}'.format(client.__class__.__name__))

        self.base_path = base_path
        self.repository_api = repository_api
        self.client = client

    def all(self, queryMap=None):
        """
        Gets info about all libraries which belong to this user.

        Not complete information is provided by all(). To get detailed information about experiment use get().

        :return: info about libraries
        :rtype: list[ExperimentsArtifact]
        """
        all_libraries = self.repository_api.v3_libraries_list(queryMap)
        list_libraries_artifact = []
        if all_libraries is not None:
            resr = all_libraries.resources

            for iter1 in resr:
                list_libraries_artifact.append(WmlLibrariesAdapter(iter1, self.client).artifact())
            return list_libraries_artifact
        else:
            return []

    def get(self, libraries_id):

        """
        Gets detailed information about libraries.

        :param str libraries_id: uid used to identify experiment
        :return: returned object has all attributes of SparkPipelineArtifact but its class name is PipelineArtifact
        :rtype: PipelineArtifact(SparkPipelineLoader)
        """

        if not isinstance(libraries_id, str) and not isinstance(libraries_id, unicode):
            raise ValueError('Invalid type for experiment_id: {}'.format(libraries_id.__class__.__name__))
        if(libraries_id.__contains__("/v4/libraries")):
            matched = re.search('.*/v4/libraries/([A-Za-z0-9\-]+)', libraries_id)
            if matched is not None:
                library_id = matched.group(1)
                return self.get(library_id)
            else:
                raise ValueError('Unexpected artifact href: {} format'.format(libraries_id))
        else:
            libraries_output = self.repository_api.v3_libraries_get(libraries_id)
            if libraries_output is not None:
                return WmlLibrariesAdapter(libraries_output, self.client).artifact()
            else:
                raise Exception('Library not found'.format(libraries))

    def remove(self, libraries_id):
        """
        Removes libraries with given libraries_id.

        :param str libraries_id: uid used to identify experiment
        """

        if not isinstance(libraries_id, str) and not isinstance(libraries_id, unicode):
            raise ValueError('Invalid type for libraries_id: {}'.format(libraries_id.__class__.__name__))

        if(libraries_id.__contains__("/v4/libraries")):
            matched = re.search('.*/v4/libraries/([A-Za-z0-9\-]+)', libraries_id)
            if matched is not None:
                libraries_id_value = matched.group(1)
                self.remove(libraries_id_value)
            else:
                raise ValueError('Unexpected experiment artifact href: {} format'.format(libraries_id))
        else:
            return self.repository_api.v3_libraries_delete_id(libraries_id)

    def patch(self, libraries_id, artifact):
        libraries_patch_input = self.prepare_libraries_patch_input(artifact)
        libraries_patch_output = self.repository_api.v3_libraries_patch_with_http_info(libraries_patch_input, libraries_id)
        statuscode = libraries_patch_output[1]

        if statuscode != 200:
            logger.info('Error while patching libraries: no location header')
            raise ApiException(statuscode,"Error while patching libraries")

        if libraries_patch_output is not None:
            new_artifact = WmlLibrariesAdapter(libraries_patch_output[0], self.client).artifact()
        return new_artifact

    def save(self, artifact, library = None):
        """
        Saves libraries in repository service.

        :param artifact : LibrariesArtifact to be saved in the repository service
        :param library : library file path to be saved in the repository service
        :return: saved artifact with changed MetaProps
        :rtype: LibrariesArtifact
        """
        logger.debug('Creating a new WML Libraries : {}'.format(artifact.name))

        if not issubclass(type(artifact), LibrariesArtifact):
            raise ValueError('Invalid type for artifact: {}'.format(artifact.__class__.__name__))

        library_input = self._prepare_wml_library_input(artifact)
        library_output = self.repository_api.v3_libraries_create_with_http_info(library_input)

        library_id = library_output[0].metadata.guid

        if library is not None:
            if not os.path.exists(library):
                raise IOError('The library specified ( {} ) does not exist.'.format(library))
            artifact.library = library

        if library is None and artifact.library is not None:
            if not os.path.exists(artifact.library):
                raise IOError('The artifact specified ( {} ) does not exist.'.format(artifact.library))

        statuscode = library_output[1]
        if statuscode != 201:
            logger.info('Error while creating libraries: no location header')
            raise ApiException(statuscode, 'No artifact location')

        if library_output is not None:
            new_artifact = WmlLibrariesAdapter(library_output[0], self.client).artifact()
            #new_artifact = WmlLibrariesAdapter(library_output, self.client).artifact()
            new_artifact.library = artifact.library
            self._upload_libraries_content(new_artifact, artifact.library)
        return new_artifact


    @staticmethod
    def _prepare_wml_library_input(artifact):
        name = None
        version = None
        platform =None
        description = None

        name = artifact.meta.prop(MetaNames.LIBRARIES.NAME)
        version = artifact.meta.prop(MetaNames.LIBRARIES.VERSION)
        description = artifact.meta.prop(MetaNames.LIBRARIES.DESCRIPTION)
        if description is not None and not isinstance(description, str):
            raise ValueError('Invalid data format for description.')

        platform = json.loads(artifact.meta.prop(MetaNames.LIBRARIES.PLATFORM))
        platform_version = platform.get('versions')
        if platform_version is not None:
            if not issubclass (type(platform_version), list):
                raise ValueError('Invalid data format for platform.version.')
        platform_input = LibrariesDefinitionInputPlatform(platform.get('name'), platform.get('versions'))

        library_input = LibrariesDefinitionInput(
            name, description, version, platform_input)
        return library_input

    @staticmethod
    def prepare_libraries_patch_input(artifact):
        patch_list = []
        patch_input = artifact.meta.prop(MetaNames.LIBRARIES.PATCH_INPUT)
        if isinstance(patch_input, str):
            patch_input_list = json.loads(artifact.meta.prop(MetaNames.LIBRARIES.PATCH_INPUT))
            if isinstance(patch_input_list, list):
                for iter1 in patch_input_list:
                    libraries_patch = PatchOperationLibraries(
                        op = iter1.get('op'),
                        path= iter1.get('path'),
                        value = iter1.get('value', None),
                        _from =iter1.get('from', None),
                    )
                    patch_list.append(libraries_patch)

                return patch_list

    def _upload_libraries_content(self, libraries_artifact, artifact_path, query_param=None):
        lib_id = libraries_artifact.uid
        content_stream = libraries_artifact.reader().read()
        self.repository_api.upload_libraries(lib_id, content_stream)
        content_stream.close()
        libraries_artifact.reader().close()
        logger.debug('Content uploaded for libraries created at: {}'.format(lib_id))




#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2021- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

import argparse
import logging
import os
import re
import sys
import time
from zipfile import ZipFile

fl_path = os.path.abspath(".")
if fl_path not in sys.path:
    sys.path.append(fl_path)

import ibmfl.envs as fl_envs
import ibmfl.util.config as fl_config
from ibmfl._version import __version__
from ibmfl.connection.route_declarations import get_party_router
from ibmfl.connection.router_handler import Router
from ibmfl.exceptions import FLException, InvalidConfigurationException
from ibmfl.message.message import Message
from ibmfl.message.message_type import MessageType
from ibmfl.util.config import (
    configure_logging_from_file,
    convert_bytes_to_zip,
    get_class_by_name,
    get_lt_from_config,
    get_party_config,
    read_yaml_config,
)

logger = logging.getLogger(__name__)


class Party:
    """
    Party participates in Federated Learning training.
    """

    def __init__(self, **kwargs):
        """
        Initializes a party based on the configuration provided in
        either a yaml file or a dictionary.
        :param config_file: path to yml file with the configuration of the party
        :type config_file: `str`
        :param config_dict: dictionary containing the configuration of the party
        :type config_dict: `dict`
        :return None
        """
        log_level = kwargs.get("log_level", "INFO")
        configure_logging_from_file(log_level=log_level)

        config_file = kwargs.get("config_file")
        config_dict = kwargs.get("config_dict")
        if config_file != None:
            config_dict = fl_config.read_yaml_config(config_file=config_file)

        config_dict["connection"]["name"] = "SWSConnection"
        config_dict["connection"]["path"] = "ibmfl.connection.websockets_connection"
        # Remove and save possible incomplete local training to be set later with agg config
        self.local_training = config_dict.pop("local_training", {})
        self.local_training_handler = None

        cls_config = fl_config.get_cls_by_config(config_dict)

        self.party_config = cls_config

        self.data_handler = None
        self.fl_model = None
        self.is_running = False

        data_config = cls_config.get("data")
        connection_config = cls_config.get("connection")
        ph_config = cls_config.get("protocol_handler")
        mrec_config = cls_config.get("metrics_recorder")

        try:
            # Load data (optional field)
            # - in some cases the aggregator doesn't have data for testing purposes
            data_cls_ref = data_config.get("cls_ref")
            data_info = data_config.get("info")
            self.data_handler = data_cls_ref(data_config=data_info)
            # Read and create model (optional field)
            # In some cases aggregator doesn't need to load the model:

            self.fl_model = None

            # Load hyperparams
            self.hyperparams = cls_config.get("hyperparams")
            self.agg_info = cls_config.get("aggregator")
            self.agg_info["ip"] = self.agg_info["ip"] + "/aggregator"
            self.agg_info["port"] = 443

            connection_cls_ref = connection_config.get("cls_ref")
            self.connection_info = connection_config.get("info")
            self.connection_info["VERSION"] = __version__
            self.connection_info["ip"] = self.agg_info["ip"]
            self.connection_info["port"] = self.agg_info["port"]
            connection_synch = connection_config.get("sync")
            # Use opt-in privacy for WML, to get metrics by default
            connection_private = self.connection_info.get("private", False)
            self.connection = connection_cls_ref(self.connection_info)
            self.connection.initialize_sender()

            ph_cls_ref = ph_config.get("cls_ref")
            if mrec_config:
                mrec_cls_ref = mrec_config.get("cls_ref")
                metrics_recorder = mrec_cls_ref(
                    mrec_config.get("output_file"),
                    mrec_config.get("output_type"),
                    mrec_config.get("compute_pre_train_eval"),
                    mrec_config.get("compute_post_train_eval"),
                )
            else:
                metrics_recorder = None

            self.proto_handler = ph_cls_ref(
                self.fl_model,
                self.connection.sender,
                self.data_handler,
                self.local_training_handler,
                metrics_recorder,
                agg_info=self.agg_info,
                synch=connection_synch,
                is_private=connection_private,
            )

            self.router = Router()
            get_party_router(self.router, self.proto_handler)

            # check for token
            token = kwargs.get("token", None)
            self_signed_cert_flag = kwargs.get("self_signed_cert", None)
            # self.connection.initialize(router=self.router, aggIinfo=self.agg_info, ssl_context=self_signed_cert_flag, authToken=token)
            self.connection.initialize_receiver(router=self.router, verify=not self_signed_cert_flag, token=token)

            self.connection.start()

        except Exception as ex:
            logger.info("Error occurred " "while loading aggregator configuration")
            logger.exception(ex)
        else:
            logger.info("Party initialization successful")

        self.agg_info = cls_config.get("aggregator")

    def initialize_model_config(self):
        logger.info("Initializing model configuration")
        cls_config = self.party_config
        model_config = cls_config.get("model")
        lt_config = cls_config.get("local_training")
        try:
            # Read and create model (optional field)
            # In some cases aggregator doesn't need to load the model:
            model_cls_ref = model_config.get("cls_ref")
            spec = model_config.get("spec")
            info = model_config.get("info")
            self.fl_model = model_cls_ref("", spec, info=info)

        except Exception as ex:
            logger.error("Error occurred while loading model config. Please check for missing libraries.")
            logger.exception(ex)
            raise InvalidConfigurationException("Error occurred while loading model config.")

        try:
            self.hyperparams = cls_config.get("hyperparams")

            lt_cls_ref = lt_config.get("cls_ref")
            lt_info = lt_config.get("info")
            self.local_training_handler = lt_cls_ref(
                self.fl_model, self.data_handler, hyperparams=self.hyperparams, info=lt_info
            )

            self.proto_handler.set_model(self.fl_model)
            self.proto_handler.set_training_handler(self.local_training_handler)

        except Exception as ex:
            logger.info("Error occurred " "while loading model and lth configuration")
            logger.exception(ex)
        else:
            logger.info("Party model initialization successful")

    def register_party(self):
        """
        Registers party with the aggregator.

        :param: None
        :return: None
        """
        logger.info("Registering party...")
        returnValue = False
        register_message = Message(MessageType.REGISTER.value, data=self.connection_info)
        try:
            response = self.connection.sender.send_message(self.agg_info, register_message)
            if response is not None and response.get_data()["status"] == "success":
                data = response.get_data()
                if self.extract_config_from_stream(data):
                    logger.info("Registration Successful")
                    returnValue = True
                else:
                    logger.info("Registration Failed: Model processing error")
                    returnValue = False
            else:
                logger.error("Registration Failed: Failure status from aggregator")
                if response is not None:
                    logger.error(response.get_data().get("detail", "No Detail Provided"))
        except Exception as ex:
            logger.error("Error occurred during registration" + str(ex))

        return returnValue

    def stop_connection(self):
        """
        Stop the connection to the aggregator

        :param: None
        :return: None
        """
        try:
            self.connection.stop()

        except Exception as ex:
            logger.error("Error occurred during stop")
            logger.error(ex)
        else:
            logger.info("Party stop successful")

    def evaluate_model(self):
        """
        Calls function that evaluates current model with local testing data
        and prints the results.

        :param: None
        :return: None
        """
        try:
            self.proto_handler.print_evaluate_local_model()

        except Exception as ex:
            logger.error("Error occurred during evaluation.")
            logger.error(ex)

    def start(self):
        """
        Initializes connection and registers with the aggregator,
        then accept commands from the aggregator to effect training.

        :param: None
        :return: None
        """

        if not self.is_running:
            logger.info("Party not registered yet.")
            if self.register_party():
                logger.info("Listening for commands from Aggregator")
                self.is_running = True
                self.initialize_model_config()
            else:
                self.stop_connection()
                raise FLException("Error occurred while registering party. ")
        else:
            logger.info("Party already running.")

    def extract_config_from_stream(self, data):
        """Read model from response stream and extract the content.
        :param data: response data recieved from aggregator.
        :type data: `json`
        :return: None
        """
        returnValue = False
        working_dir = fl_envs.working_directory

        if "model_package" in data:
            logger.info("Model Package found, placing in working_dir " + working_dir)
            model_file_name = "model_package.zip"
            model_file_op = os.path.join(working_dir, model_file_name)
            model_pk_bytes = data.get("model_package")

            try:
                convert_bytes_to_zip(model_pk_bytes, model_file_op)

                with ZipFile(model_file_op, "r") as o:
                    o.extractall(working_dir)

            except Exception as ex:
                logger.exception("Error occurred while unpacking model package from aggregator.")
                raise FLException("Error occurred while downloading model package. ")
        else:
            logger.info("No model package received from aggregator.")

        logger.info("Checking for model config in payload.")

        model_config_agg = data.get("model_config", None)
        if model_config_agg == None:
            logger.info("No model config received from aggregator.")
            return returnValue

        if "is_model_working_dir" in model_config_agg.get("spec"):
            model_config_agg["spec"]["model_definition"] = working_dir

        elif "model_definition" in model_config_agg.get("spec"):
            model_def = os.path.join(working_dir, model_config_agg["spec"]["model_definition"])
            model_config_agg["spec"]["model_definition"] = model_def

        logger.info("Aggregator model config: %s" % str(model_config_agg))
        model_config_local = self.party_config.get("model") or {}
        logger.info("Local model config: %s" % str(model_config_local))
        model_config_agg.update(model_config_local)
        logger.info("Merged model config: %s" % str(model_config_agg))

        self.party_config["model"] = model_config_agg

        local_training_agg = data.get("local_training")
        logger.info("Default local training: %s" % str(local_training_agg))
        logger.info("Configured local training: %s" % str(self.local_training))

        # Merge local trainging config:
        # name and path may be overridden:
        if "name" in self.local_training:
            local_training_agg["name"] = self.local_training["name"]
        if "path" in self.local_training:
            local_training_agg["path"] = self.local_training["path"]
        # Crypto key_mgr_info may be overridden
        if (
            "info" in self.local_training
            and ("crypto" in self.local_training["info"])
            and ("key_manager" in self.local_training["info"]["crypto"])
            and ("key_mgr_info" in self.local_training["info"]["crypto"]["key_manager"])
            and ("info" in local_training_agg)
            and ("crypto" in local_training_agg["info"])
            and ("key_manager" in local_training_agg["info"]["crypto"])
        ):
            local_training_agg["info"]["crypto"]["key_manager"]["key_mgr_info"] = self.local_training["info"]["crypto"][
                "key_manager"
            ]["key_mgr_info"]

        logger.info("Merged local training: %s" % str(local_training_agg))

        if local_training_agg == None:
            logger.error("No local training config received from aggregator.")
            return returnValue
        self.party_config["local_training"] = get_lt_from_config(local_training_agg)

        logger.info("config extract finished!")

        returnValue = True
        return returnValue


if __name__ == "__main__":
    """
    Main function to run a party-side application in a Federated Learning training job.
    The application can either be run in an interactive ( -i ), or a non-interactive
    mode.

    In interactive mode, enter commands:

    START     - Start the connection for communication with the aggregator
    STOP      - Stop the connection
    REGISTER  - Register with the aggregator and start accepting commands
    EVAL      - Evaluate the model with local test data

    In non-interactive mode, the party will be started, and registered with
    the aggregator.  After training is complete and a STOP message is received
    from the aggregator, the application will exit.

    """

    parser = argparse.ArgumentParser(description="WML Federated Learning Party")
    parser.add_argument("config_file", help="yaml configuration file")
    parser.add_argument("token", help="authentication token")
    parser.add_argument("-s", "--self_signed_cert", help="flag for self-signed certificate", action="store_true")
    parser.add_argument("-i", "--interactive", help="run interactively", action="store_true")
    parser.add_argument(
        "-l",
        "--log_level",
        type=str,
        default="INFO",
        required=False,
        help="log_level should be a value from [DEBUG, INFO, WARNING, ERROR, CRITICAL]",
    )
    args = parser.parse_args()

    if args.self_signed_cert:
        self_signed_cert_arg = True
    else:
        self_signed_cert_arg = None

    p = Party(
        config_file=args.config_file, token=args.token, self_signed_cert=self_signed_cert_arg, log_level=args.log_level
    )

    if args.interactive:
        # Indefinite loop to accept user commands to execute
        while 1:
            try:
                msg = sys.stdin.readline()
                if re.match("START", msg):
                    p.start()

                if re.match("STOP", msg):
                    p.stop_connection()
                    break

                if re.match("REGISTER", msg):
                    p.register_party()

                if re.match("EVAL", msg):
                    p.evaluate_model()

            except Exception as ex:
                logger.error("Error occurred during " + msg)
                logger.error(ex)
    else:
        try:
            p.start()
        except Exception as ex:
            logger.error("Error occurred during start")
            logger.error(ex)
            p.stop_connection()

        while not p.connection.stopped:
            logger.debug("Party is still running")
            time.sleep(10)
        logger.info("Party is complete")



#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2020- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

from abc import ABC

__all__ = [
    "BaseLocation"
]


class BaseLocation(ABC):
    """Base class for storage Location."""
    def to_dict(self) -> dict:
        """Get a json dictionary representing this model."""
        return vars(self)

    def _get_file_size(self, **kwargs) -> 'int':
        raise NotImplementedError



#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2018- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

from ibm_watson_machine_learning.libs.repo.mlrepository import MetaNames
from ibm_watson_machine_learning.libs.repo.mlrepository import MetaProps
from ibm_watson_machine_learning.libs.repo.mlrepository import ModelArtifact
from .python_version import PythonVersion
from ibm_watson_machine_learning.libs.repo.util.unique_id_gen import uid_generate

import os, shutil, tarfile, json


class HybridPipelineModelArtifact(ModelArtifact):
    """
    Class of Hybrid model artifacts created with MLRepositoryCLient.

    """
    def __init__(self, hybrid_pipeline_model=None, uid=None, name=None, meta_props=MetaProps({})):
        super(HybridPipelineModelArtifact, self).__init__(uid, name, meta_props)

        self.ml_pipeline_model = hybrid_pipeline_model
        self.ml_pipeline = None     # no pipeline or parent reference

        if meta_props.prop(MetaNames.RUNTIMES) is None and meta_props.prop(MetaNames.RUNTIME_UID) is None and meta_props.prop(MetaNames.FRAMEWORK_RUNTIMES) is None:
            ver = PythonVersion.significant()
            runtimes = '[{"name":"python","version": "'+ ver + '"}]'
            self.meta.merge(
                MetaProps({MetaNames.FRAMEWORK_RUNTIMES: runtimes})
            )

    def pipeline_artifact(self):
        """
        Returns None. Pipeline is not implemented for Tensorflow model.
        """
        pass

    def _copy(self, uid=None, meta_props=None):
        if uid is None:
            uid = self.uid

        if meta_props is None:
            meta_props = self.meta

        return HybridPipelineModelArtifact(
            self.ml_pipeline_model,
            uid=uid,
            name=self.name,
            meta_props=meta_props
        )





#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2017- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

from .spark_version import SparkVersion

from ibm_watson_machine_learning.libs.repo.util.library_imports import LibraryChecker
from ibm_watson_machine_learning.libs.repo.base_constants import *

lib_checker = LibraryChecker()

if lib_checker.installed_libs[SCIKIT]:
    from sklearn.base import BaseEstimator

class VersionHelper(object):
    @staticmethod
    def significant():
        import pipeline
        version_parts = [int(version_part) for version_part in pipeline.__version__.split('.')]
        return "{}.{}".format(version_parts[0], version_parts[1])

    @staticmethod
    def getFrameworkVersion(ml_pipeline):
        canonical_name = ml_pipeline.__class__.__name__
        if canonical_name == 'Pipeline':
            return SparkVersion.significant()
        elif canonical_name =='PipelineModel':
            return SparkVersion.significant()
        elif canonical_name == 'MLPipeline':
            return '0.1'
        elif canonical_name =='MLPipelineModel':
            return '0.1'

    @staticmethod
    def pipeline_type(ml_pipeline):
        canonical_name = ml_pipeline.__class__.__name__
        if canonical_name == 'Pipeline':
            return 'mllib'
        elif canonical_name == 'MLPipeline':
            return 'wml'
        else:
            raise ValueError('Unsupported Pipeline class: {}'.format(canonical_name))

    @staticmethod
    def model_type(ml_pipeline_model):
        class_name = ml_pipeline_model.__class__.__name__
        if class_name == 'PipelineModel':
            return 'mllib'
        elif class_name == 'MLPipelineModel':
            return 'wml'
        else:
            raise ValueError('Unsupported PipelineModel class: {}'.format(class_name))


class ScikitVersionHelper(object):
    @staticmethod
    def model_type(scikit_pipeline_model):
        lib_checker.check_lib(SCIKIT)
        class_name = scikit_pipeline_model.__class__.__name__
        if issubclass(type(scikit_pipeline_model), BaseEstimator):
            return 'scikit-learn'
        else:
            raise ValueError('Unsupported ScikitPipelineModel class: {}'.format(class_name))

    @staticmethod
    def model_version(scikit_pipeline_model):
        class_name = scikit_pipeline_model.__class__.__name__
        if issubclass(type(scikit_pipeline_model), BaseEstimator):
            import sklearn
            sk_version = sklearn.__version__
            if len(sk_version.split('.'))>2:
                split_sk = sk_version.split('.')
                sk_version = split_sk[0]+'.' + split_sk[1]
            return format(sk_version)
        else:
            raise ValueError('Unsupported ScikitPipelineModel class: {}'.format(class_name))


class TensorflowVersionHelper(object):

    @staticmethod
    def model_type(tensorflow_pipeline_model):
        lib_checker.check_lib(TENSORFLOW)
        class_name = tensorflow_pipeline_model.__class__.__name__
        import tensorflow as tf
        if '2.1.0' in tf.__version__ :
            if isinstance(tensorflow_pipeline_model, tf.compat.v1.Session):
                return 'tensorflow'
            else:
                raise TypeError(
                    "Expecting object of type : %s" % tf.compat.v1.Session + " but got %s" % type(tensorflow_pipeline_model))

        else:
            if isinstance(tensorflow_pipeline_model, tf.Session):
                return 'tensorflow'
            else:
                raise TypeError("Expecting object of type : %s" % tf.Session + " but got %s" % type(tensorflow_pipeline_model))



    @staticmethod
    def model_version(tensorflow_pipeline_model):
        import tensorflow as tf
        class_name = tensorflow_pipeline_model.__class__.__name__
        if '2.1.0' in tf.__version__ :
            if isinstance(tensorflow_pipeline_model, tf.compat.v1.Session):
                tf_version = tf.__version__
                if len(tf_version.split('.')) > 2:
                    split_tf = tf_version.split('.')
                    tf_version = split_tf[0] + '.' + split_tf[1]
                return format(tf_version)
            else:
                raise TypeError(
                    "Expecting object of type : %s" % tf.compat.v1.Session + " but got %s" % type(tensorflow_pipeline_model))
        else:

            if isinstance(tensorflow_pipeline_model,tf.Session):
                tf_version = tf.__version__
                if len(tf_version.split('.'))>2:
                    split_tf = tf_version.split('.')
                    tf_version = split_tf[0]+'.' + split_tf[1]
                return format(tf_version)
            else:
                raise TypeError("Expecting object of type : %s" % tf.Session + " but got %s" % type(tensorflow_pipeline_model))



class XGBoostVersionHelper(object):
    @staticmethod
    def model_type(xgboost_model):
        lib_checker.check_lib(XGBOOST)
        from xgboost import Booster
        if isinstance(xgboost_model, Booster):
            return 'xgboost'
        else:
            raise ValueError('Unsupported XGBoost model class: {}'.format(xgboost_model.__class__.__name__))


    @staticmethod
    def model_version(xgboost_model):
        lib_checker.check_lib(XGBOOST)
        from xgboost import Booster
        import xgboost
        if isinstance(xgboost_model, Booster):
            xgb_version = xgboost.__version__
            return xgb_version

class ScikitModelBinary(object):
    @staticmethod
    def model_bin_name():
        return "scikit_model.pkl"

    @staticmethod
    def bin_ext():
        return ".pkl"

    @staticmethod
    def bin_ext_v0():
        return ".bin"

class XGBoostModelBinary(object):
    @staticmethod
    def model_bin_name():
        return "xgboost_model.pkl"

    @staticmethod
    def bin_ext():
        return ".pkl"

class TensorflowModelBinary(object):

    @staticmethod
    def bin_ext():
        return ".pb"
